
import jahuwaldt.io.ExtFilenameFilter;
import jahuwaldt.io.FileUtils;
import jahuwaldt.js.datareader.DataCase;
import jahuwaldt.js.datareader.DataReader;
import jahuwaldt.js.datareader.DataReaderFactory;
import jahuwaldt.js.datareader.DataSet;
import jahuwaldt.swing.*;
import java.awt.*;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.io.*;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.GZIPInputStream;
import javax.swing.*;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * Main window for the qPlot program. Most of the program code is based here.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt    Date: September 3, 2011
 * @version January 15, 2024
 */
@SuppressWarnings("serial")
public class MainWindow extends JFrame {

    //  The logger for this class.
    private static final Logger LOGGER = Logger.getLogger(MainWindow.class.getName());
    
    //  A count of how many windows have been opened.
    private static int _windowCount = 0;
    
	// The list of runs shown in the GUI.
	private JList runList = null;

	// The data model for the runList.
	private DefaultListModel listModel = new DefaultListModel();
	
	//  A database to hold our data.
	private DataSet _data;
	
	//	A list of optional plot definitions.
	private PlotDefinitionList _plotDefs = null;
	
    //  A reference to the Reload button.
    private JButton reloadButton;
	
	/**
	*  Constructor for our application window.
	*
	*  @param  newData     The data set this window contains.
	*  @param  application A reference to the application that created this window.
	*/
    @SuppressWarnings({"OverridableMethodCallInConstructor", "LeakingThisInConstructor"})
	private MainWindow(DataSet newData) throws NoSuchMethodException {
        super(MessageFormat.format(MDIApplication.getInstance().getResourceBundle().getString("mainTitle"), (_windowCount > 0 ? String.valueOf(_windowCount) : "")));
		
		//  Have the window dispose of itself when it closes.
		this.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		
        //  Define the window icon in the taskbar.
        ImageIcon icon = AppUtilities.getResourceIcon(MDIApplication.getInstance().getResourceBundle().getString("applicationIconResName"), LOGGER);
        if (icon != null)
            this.setIconImage(icon.getImage());

		//  Save a reference to the new data set.
		_data = newData;
		
		// Layout the display of this window.
		layoutDisplay(newData);
		
		//	Set up the menu bar.
		JMenuBar menuBar = createMenuBar();
		this.setJMenuBar(menuBar);

		//	Register a quit listener so that the user can save unsaved changes before the application quits.
        MDIApplication.addQuitListener(new QuitListener() {
            @Override
            public boolean quit() {
                return handleQuit();
            }
        });
        
		//  Size the window.
		this.pack();
		
		//  Position the window so that each new window can be seen.
		AppUtilities.positionWindow( this, getWidth(), getHeight() );
        
        //  Increment the window count.
        ++_windowCount;
	}
	

	//-----------------------------------------------------------------------------------
	/**
	*  Sets the title for this frame to the specified string.  Also notifies the main
	*  application class so that the "Windows" menu gets updated too.
	*
	*  @param  title The title to be displayed in the frame's border.
	*                A null value is treated as an empty string, "".
	*/
    @Override
	public void setTitle(String title) {
		super.setTitle(title);
		MDIApplication.windowTitleChanged(this);
	}
	
	/**
	*  Layouts the contents of this main application window.
	*
	*  @param  theData  The time line data being plotted.
	*/
	private void layoutDisplay(DataSet theData) throws NoSuchMethodException {
		ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
		
		// Layout the main application window's user interface.
		JLabel selectLabel = new JLabel(resBundle.getString("selectRunLabel"), JLabel.LEFT);
		selectLabel.setAlignmentX(JLabel.LEFT_ALIGNMENT);

		// Set up buttons.
		JButton removeButton = new JButton( resBundle.getString("removeBtnText") );
		removeButton.addActionListener( AppUtilities.getActionListenerForMethod(this, "handleRemove") );
		removeButton.setCursor( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR ) );
		removeButton.setToolTipText( resBundle.getString("removeBtnToolTip") );
		
		JButton addButton = new JButton( resBundle.getString("addBtnText") );
		addButton.addActionListener( AppUtilities.getActionListenerForMethod(this, "handleAdd") );
		addButton.setCursor( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR ) );
		addButton.setToolTipText( resBundle.getString("addBtnToolTip") );
		
		reloadButton = new JButton( resBundle.getString("reloadBtnText") );
		reloadButton.addActionListener( AppUtilities.getActionListenerForMethod(this, "handleReload") );
		reloadButton.setCursor( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR ) );
		reloadButton.setToolTipText( resBundle.getString("reloadBtnToolTip") );
        reloadButton.setEnabled(false);
		
		JPanel addRemoveBtnPanel = new JPanel( new GridLayout( 3, 1, 15, 15 ) );
		addRemoveBtnPanel.setBorder( new EmptyBorder( 20, 10, 20, 10 ) );
		addRemoveBtnPanel.add( addButton );
		addRemoveBtnPanel.add( removeButton );
        addRemoveBtnPanel.add( reloadButton );

		JButton plotButton = new JButton( resBundle.getString("plotBtnText") );
		plotButton.addActionListener( AppUtilities.getActionListenerForMethod(this, "handlePlot") );
		plotButton.setCursor( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR ) );
		plotButton.setToolTipText( resBundle.getString("plotBtnToolTip") );
		
		JPanel plotBtnPanel = new JPanel( new FlowLayout( FlowLayout.CENTER, 10, 5 ) );
		plotBtnPanel.add( plotButton );
		
		// Set up the run list.
		Box listBox = Box.createVerticalBox();
		listBox.setBorder(BorderFactory.createEmptyBorder(5,20,0,0));
		listBox.add(selectLabel);
		for (DataCase aCase : theData ) {
			listModel.addElement(aCase);
		}
		
		runList = new JList( listModel );
		runList.setPrototypeCellValue( "A Typical Run Name Entry Size" );
		runList.setSelectionMode( ListSelectionModel.MULTIPLE_INTERVAL_SELECTION );
        runList.addListSelectionListener(new ListSelectionListener() {
            @Override
            public void valueChanged(ListSelectionEvent e) {
                JList lst = (JList)e.getSource();
                if (!e.getValueIsAdjusting()) {
                    if (lst.isSelectionEmpty()) {
                        reloadButton.setEnabled(false);
                    } else {
                        reloadButton.setEnabled(true);
                    }
                }
            }
            
        });
		JScrollPane scrollPane = new JScrollPane( runList );
		scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
		scrollPane.setAlignmentX(JScrollPane.LEFT_ALIGNMENT);
		listBox.add(scrollPane);
		
		// Set up application information.
		Box appInfoPanel = Box.createHorizontalBox();
		appInfoPanel.setBorder( new EmptyBorder( 5, 20, 0, 10 ) );
		JLabel appName = new JLabel( resBundle.getString("appName") );
		Font font = appName.getFont().deriveFont(Font.BOLD);
		font = font.deriveFont(font.getSize2D()*1.2F);
		appName.setFont(font);
		JLabel version = new JLabel( resBundle.getString("version") + " " + resBundle.getString("appVersion") );
		Box vbox = Box.createVerticalBox();
		vbox.add( appName );
		vbox.add( version );
		appInfoPanel.add(vbox);		
		
		// Layout this window's content pane.
		Container cp = this.getContentPane();
		cp.add( appInfoPanel, BorderLayout.NORTH );
		cp.add( addRemoveBtnPanel, BorderLayout.EAST );
		cp.add( plotBtnPanel, BorderLayout.SOUTH );
		cp.add( listBox, BorderLayout.CENTER );
		
		
		//  Add support for drag and drop to the run list.
		new DropTarget(runList, DnDConstants.ACTION_COPY, new FileDTListener(), true);
	}
	
	/**
	*  A class that listens for Drag & Drop events that contain files.  When a drop event is received,
	*  this class reads in any files that were dropped.
	*/
	private class FileDTListener extends DropTargetAdapter {
		
		/**
		*  Called when a drag operation is ongoing, while the mouse pointer is still
		*  over the operable part of the drop site for the <code>DropTarget</code>
		*  registered with this listener.
		*  This implementation will accept only COPY drop actions for javaFileListFlavor flavors.
		*/
        @Override
		public void dragOver(DropTargetDragEvent ev)  {
			dropTargetDrag(ev);
		}
		
		/**
		*  Called if the user has modified the current drop gesture.
		*  This implementation will accept only COPY drop actions for javaFileListFlavor flavors.
		*/
        @Override
		public void dropActionChanged(DropTargetDragEvent ev) {
			dropTargetDrag(ev);
		}
		
		private void dropTargetDrag(DropTargetDragEvent ev) {
			//	Accept only file list compatible data.
			if (ev.isDataFlavorSupported(DataFlavor.javaFileListFlavor))
				ev.acceptDrag(DnDConstants.ACTION_COPY);	//	Accept only a COPY type of transfer.
			else
				ev.rejectDrag();
		}
		
		/**
		* Called when the drag operation has terminated with a drop on the
		* operable part of the drop site for the <code>DropTarget</code> registered
		* with this listener.
		* This implementation will process the drop event, extract the file list
		* and load each file in the list.
		*/
		@SuppressWarnings("unchecked")
        @Override
		public void drop(DropTargetDropEvent ev) {
			
			//	Accept only COPY drops.
			ev.acceptDrop(DnDConstants.ACTION_COPY);

			try {
				ev.getSource();
				List<File> files = (List<File>)ev.getTransferable().getTransferData(DataFlavor.javaFileListFlavor);
				
				//	Process the files, one at a time.
				int oldSize = _data.size();
				int addedCount = 0;
				for (File theFile : files) {
					
					//	Is the file a real file (and not a directory)?
					if (theFile.isFile()) {
						//	Try to read in this file.
                        addedCount += openFile(theFile);
					}
				}
				
				//	Select the runs that were added.
				if (addedCount > 0)
					runList.setSelectionInterval(oldSize, oldSize+addedCount-1);
				
			} catch (UnsupportedFlavorException | IOException ex) {
				LOGGER.log(Level.WARNING, null, ex);
			}
			
            ev.dropComplete(true);
            }
		}
	
	
	/**
	*  Initializes the menus associated with this window.
	*/
	private JMenuBar createMenuBar() throws NoSuchMethodException {
		ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
		
		JMenuBar menuBar = new JMenuBar();
		
		// Set up the file menu.
		int row=0;
		String[][] menuStrings = new String[10][3];
		menuStrings[row][0] = resBundle.getString("newItemText");
		menuStrings[row][1] = resBundle.getString("newItemKey");
		menuStrings[row][2] = "handleNew";
		++row;
		menuStrings[row][0] = resBundle.getString("openItemText");
		menuStrings[row][1] = resBundle.getString("openItemKey");
		menuStrings[row][2] = "handleOpen";
		++row;
		menuStrings[row][0] = resBundle.getString("openPlotDefsItemText");
		menuStrings[row][1] = null;
		menuStrings[row][2] = "handleOpenPlotDefs";
		++row;
		++row;	//	Blank line
		menuStrings[row][0] = resBundle.getString("closeItemText");
		menuStrings[row][1] = resBundle.getString("closeItemKey");
		menuStrings[row][2] = "handleClose";
		++row;
		menuStrings[row][0] = resBundle.getString("exportAsItemText");
		menuStrings[row][1] = null;
		menuStrings[row][2] = null;
		int exportAsItem = row;			//	Store the iten index of the Export menu for later use.
		++row;
		++row;	//	Blank line
		menuStrings[row][0] = resBundle.getString("pageSetupItemText");
		menuStrings[row][1] = null;
		menuStrings[row][2] = null;
		++row;
		menuStrings[row][0] = resBundle.getString("printItemText");
		menuStrings[row][1] = resBundle.getString("printItemKey");
		menuStrings[row][2] = null;
		JMenu menu = AppUtilities.buildMenu( this, resBundle.getString("fileMenuText"), menuStrings );
		menuBar.add( menu );
        
		//	Add the Export As sub-menu.
		JMenu exportMenu = buildExportMenu(menu.getItem(exportAsItem).getText());
		if (nonNull(exportMenu)) {
			menu.remove(exportAsItem);
			menu.add(exportMenu, exportAsItem);
		}
        
        //  Add a Quit menu item (if it isn't automatically present).
        AppGUI guiApp = (AppGUI)MDIApplication.getInstance();
        if (!QuitJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(guiApp.getQuitJMenuItem());
        }
        
        
		// Set up the edit menu.
		row=0;
		menuStrings = new String[6][3];
		menuStrings[row][0] = resBundle.getString("undoItemText");
		menuStrings[row][1] = resBundle.getString("undoItemKey");
		menuStrings[row][2] = null;
		++row;
		menuStrings[row][0] = resBundle.getString("redoItemText");
		menuStrings[row][1] = null;
		menuStrings[row][2] = null;
		++row;
		++row;	//	Blank line.
		menuStrings[row][0] = resBundle.getString("cutItemText");
		menuStrings[row][1] = resBundle.getString("cutItemKey");
		menuStrings[row][2] = null;
		++row;
		menuStrings[row][0] = resBundle.getString("copyItemText");
		menuStrings[row][1] = resBundle.getString("copyItemKey");
		menuStrings[row][2] = null;
		++row;
		menuStrings[row][0] = resBundle.getString("pasteItemText");
		menuStrings[row][1] = resBundle.getString("pasteItemKey");
		menuStrings[row][2] = null;
		menu = AppUtilities.buildMenu( this, resBundle.getString("editMenuText"), menuStrings );
		
		menuBar.add( menu );

		//  Create a Window's menu.
		menu = MDIApplication.newWindowsMenu(resBundle.getString("windowsMenuText"));
		menuBar.add(menu);
		
        //  Create an about menu item (if not automatically present).
        if (!AboutJMenuItem.isAutomaticallyPresent()) {
            //  Create Help menu.
            menu = new JMenu(resBundle.getString("helpMenuText"));
            menuBar.add(menu);
            
            //  Add the "About" item to the Help menu.
            menu.add(guiApp.getAboutJMenuItem());
        }
        
		return menuBar;
	}

	/**
	*  Method that will build an "Export" menu containing a list of all the ModelWriter objects that
	*  are available.
	*
	*  @param  title	The title for the menu that is created.
	*  @returns A JMenu instance containing a list of all ModelWriter objects that can write to
	*           files in various formats.
	*/
	private JMenu buildExportMenu(String title) throws NoSuchMethodException {
		
		//  Get the list of data readers that are available.
		DataReader[] allReaders = DataReaderFactory.getAllReaders();
		if (isNull(allReaders) || allReaders.length < 1)
			return null;

		//	Create the menu
		JMenu menu = new JMenu(title);
		
		//  Loop over all the readers.
		boolean noneFound = true;
		for (DataReader reader : allReaders) {
			
			if (reader.canWriteData()) {
				//  Found a "writer".
				noneFound = false;
				
				//  Create a menu item for it.
				JMenuItem menuItem = new JMenuItem(reader.toString() + "...");
				menuItem.addActionListener( new ExportAsMenuListener(reader) );
				menuItem.setActionCommand( reader.toString() );
				menuItem.setEnabled( true );
				
				//  Add the menu item to the menu.
				menu.add(menuItem);
			}
		}
		
		if (noneFound)  menu = null;
		
		return menu;
	}
	
	/**
	*  Defines an action listener for our Export menu items.
	*/
	private class ExportAsMenuListener implements ActionListener {
		//  Keep a reference to the reader used by this menu.
		private final DataReader reader;
		
		public ExportAsMenuListener(DataReader reader) {
			this.reader = reader;
		}
		
        @Override
		public void actionPerformed( ActionEvent evt ) {
				handleSaveAs();
		}

		/**
		*  Handles saving the data using this menu item's DataReader instance.
		*/
		private void handleSaveAs() {
			if (_data.size() < 1)	return;
            
			ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
			
			//	Set up the file name filter.
			String extension = reader.getExtension();
			ExtFilenameFilter filter = new ExtFilenameFilter(extension, reader.toString() + " File");

			//	Build the directory and filename to prompt the user with.
			String dir = MDIApplication.getInstance().getPreferences().getLastPath();
			String fileName = resBundle.getString("untitled") + "." + extension;
			
			//	Replace any ":" characters with "_".
			fileName = fileName.replace(':', '_');
			
			//	Replace any "_ " with " ".
			fileName = fileName.replaceAll("_ ", " ");
			
			int idx = fileName.lastIndexOf(".");
			if (idx > 0)
				fileName = fileName.substring(0, idx);
			fileName += "." + extension;
			
			//	Ask the user to select a file for saving.
			File chosenFile = AppUtilities.selectFile4Save(MainWindow.this,
					MessageFormat.format(resBundle.getString("fileSaveDialogMsg"),reader.toString()),
					dir, fileName, filter, extension,
					resBundle.getString("fileExistsMsgFmt"), resBundle.getString("warningTitle"));
			if (nonNull(chosenFile)) {
				if (chosenFile.exists() && !chosenFile.canWrite()) {
					JOptionPane.showMessageDialog( MainWindow.this,
								resBundle.getString("canNotWrite2File"),
								resBundle.getString("ioErrorTitle"), JOptionPane.ERROR_MESSAGE );
					return;
				}

				//	Export the data.
				exportData(reader, chosenFile);
				
				//	Store off the path reference for later use.
				MDIApplication.getInstance().getPreferences().setLastPath(chosenFile.getParent());
			}
		}


		/**
		*  Method that exports the specified data to the specified file using
		*  the specified writer.
		*
		*  @param writer     The DataReader to use to write out the data.
		*  @param outFile    Reference to the file path where the file should be written.
		*/
		private void exportData(DataReader writer, File outFile) {
			
            try (FileOutputStream output = new FileOutputStream(outFile)) {
				//	Create a List of DataSet objects as required by the writer.
				List<DataSet> list = new ArrayList<>();
				list.add(_data);
				
				// Write out the selected tables using the specified writer.
				writer.write( output, list );
				
			} catch( IOException | ArrayIndexOutOfBoundsException ex ) {
				outFile.delete();
				ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
				JOptionPane.showMessageDialog(MainWindow.this, ex.getMessage(),
                        resBundle.getString("ioErrorTitle"),
						JOptionPane.ERROR_MESSAGE );
			}
		}
		
	}   //  end class ExportAsMenuListener
	
	
    /**
     * Handle the user choosing the "New..." from the File menu. Creates a new, blank,
     * document window.
     *
     * @param event
     * @return
     */
	public MainWindow handleNew( ActionEvent event ) {
		return (MainWindow)(MDIApplication.getInstance().handleNew(event));
	}

    /**
     * Handle the user choosing "Close" from the File menu. This implementation dispatches
     * a "Window Closing" event to this window.
     *
     * @param event
     */
	public void handleClose( ActionEvent event ) {
		this.dispatchEvent( new WindowEvent( this, WindowEvent.WINDOW_CLOSING ) );
	}

    /**
     * Handle the user clicking on the "Add" button. Reads in a data file selected by the
     * user and adds those runs to the ones already in memory (if any).
     *
     * @param event
     */
	public void handleAdd( ActionEvent event ) {
		handleOpen( event );
	}
	
    /**
     * Handle the user clicking on the "Reload" button. Re-loads the currently selected
     * data file.
     *
     * @param event
     */
    public void handleReload(ActionEvent event) {
        if (!runList.isSelectionEmpty()) {

            // Get a list of the items the user has selected.
            List selection = runList.getSelectedValuesList();

            //  Loop over all the selected rows.
            for (Object item : selection) {
                DataCase aCase = (DataCase)item;

                //  Get the file associated with this case.
                File theFile = (File)aCase.getUserObject();
                
                //	Read the file and replace the current case in the list.
                DataSet newData = readData(this, theFile, false);

                if (nonNull(newData)) {
                    //  Remove the original data.
                    _data.remove(_data.indexOf(item));
                    listModel.removeElement(item);

                    //  Add the new data.
                    for (DataCase aCase2 : newData) {
                        aCase2.setUserObject(theFile);
                        _data.add(aCase2);
                        listModel.addElement(aCase2);
                    }
                }

            }
            
			runList.clearSelection();
        }
    }
	
    /**
     * Handle the user clicking on the "Remove" button. Removes the selected runs from the
     * list of currently open runs.
     *
     * @param event
     */
	public void handleRemove( ActionEvent event ) {
	
		if ( ! runList.isSelectionEmpty() ) {
		
			// Get a list of the items the user has selected.
			List selection = runList.getSelectedValuesList();
			int length = selection.size();

			// Create a message to display to the user.
			ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
			StringBuilder buffer = new StringBuilder( "<html>" );
			buffer.append(resBundle.getString("removeConfirmText"));
			buffer.append("<br>");
			for ( int i = 0; i < length; ++i ) {
				buffer.append( selection.get(i).toString() );
				if ( i < length - 1 )
					buffer.append( ", " );
			}
			buffer.append("</html>");

			// Ask the user for confirmation.
			int result = JOptionPane.showConfirmDialog( this, buffer.toString(), resBundle.getString("warningTitle"),
			JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE );
			if ( result == JOptionPane.YES_OPTION ) {

				// The user has confirmed, so remove the selected runs from database and list.
				for ( Object item : selection ) {
					_data.remove( _data.indexOf(item) );
					listModel.removeElement(item);
				}
				
				runList.clearSelection();

				// Update the display.
				this.repaint();
			}
		}
		
	}
	
    /**
     * Handle the user clicking on the "Plot" button. Opens a plot window and displays the
     * currently selected run's data.
     *
     * @param event
     */
	public void handlePlot( ActionEvent event ) {
		ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
		try {
		
			//	If only 1 case is loaded, select it automatically.
			if (listModel.size() == 1)
				runList.setSelectedIndex(0);
			
			if ( !runList.isSelectionEmpty() ) {
				
				// Get a list of the items the user has selected.
				List selection = runList.getSelectedValuesList();
				
				//	Create a temporary data set from the selected items.
				DataSet tempSet = DataSet.newInstance("Temporary");
				for ( Object item : selection ) {
					DataCase theData = _data.get( _data.indexOf(item) );
					tempSet.add(theData);
				}
				
				//	Is this data set compatible with a multi-case plot window?
				if (tempSet.size() > 1 && PlotWindow.checkForMultiCaseCompatibility(tempSet) == 0) {
					//	Plot multitple selections in a single window.
					PlotWindow plot;
					if (isNull(_plotDefs))
						plot = new PlotWindow(_data.getName() + " " + resBundle.getString("plotText"), tempSet);
					else
						plot = new PlotDefPlotWindow(_data.getName() + " " + resBundle.getString("plotText"),
														tempSet, _plotDefs);
					plot.setVisible(true);
				
				} else {
					//	Plot multiple selections in separate windows.
					for ( DataCase theData : tempSet ) {
						// Create an instance of a plot window for each selected run.
						PlotWindow plot;
						if (isNull(_plotDefs))
							plot = new PlotWindow( theData.getName() + " " + resBundle.getString("plotText"), theData );
						else
							plot = new PlotDefPlotWindow(theData.getName() + " " + resBundle.getString("plotText"),
														theData, _plotDefs);
						plot.setVisible(true);
					}
				}
			}
			
        } catch (NoSuchMethodException ex) {
			LOGGER.log(Level.SEVERE, null, ex);
			AppUtilities.showException(null, resBundle.getString("noSuchMethodErr"),
								resBundle.getString("unexpectedMsg"), ex);
		}
	}
	
    /**
     * Handle the user choosing "Open..." from the File menu. Lets the user choose a data
     * file and open it.
     *
     * @param event
     */
	public void handleOpen( ActionEvent event ) {
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();

		try {
			String dir = MDIApplication.getInstance().getPreferences().getLastPath();
			File theFile = AppUtilities.selectFile(this, FileDialog.LOAD,
                    resBundle.getString("fileDialogLoad"), dir,
					"", MDIApplication.getFilenameFilter());
			if (isNull(theFile))	return;
			
            //  Read in the file and add it to the current window.
            openFile(theFile);
            
		} catch (Exception ex) {
            LOGGER.log(Level.SEVERE, null, ex);
			AppUtilities.showException(null, resBundle.getString("unexpectedTitle"),
								resBundle.getString("unexpectedMsg"), ex);
		}
	    
	}

    /**
     * Function that opens the specified data file and adds any cases in it to those
     * in the current window.
     * 
     * @param theFile The data file to read in.
     * @return A count of how many cases were read in from the file.
     */
    public int openFile(File theFile) {
        
        //	Read the file and add it to the current window.
        DataSet newData = readData(this, theFile, false);
        
        int addedCount = 0;
        if (nonNull(newData)) {
            for (DataCase aCase : newData) {
                aCase.setUserObject(theFile);
                _data.add(aCase);
                listModel.addElement(aCase);
            }
            _data.setName(newData.getName());
            ++addedCount;

        }
        
        return addedCount;
    }
    
    /**
     * Method that reads in the specified file and creates a new window for displaying
     * it's contents.
     *
     * @param parent Parent frame for dialogs (null is fine).
     * @param theFiles The list of files to be loaded and displayed. If null is passed,
     *                  this method will simply create an empty window.
     * @param autoPlot Automatically plot the data if true, don't if false.
     * @param defsFile Reference to a plot definitions file or null for none.
     * @return A reference to the newly created window.
     * @throws java.lang.NoSuchMethodException
     */
	public static MainWindow newWindowFromDataFile(Frame parent, List<File> theFiles,
								boolean autoPlot, File defsFile)
										throws NoSuchMethodException {
		
		if (isNull(theFiles))	theFiles = new ArrayList<>();
		if (theFiles.isEmpty())	theFiles.add(null);
		
		//	Read in the data from the input file.
		DataSet newData = readData(parent, theFiles.get(0), false);
		
		// Create an instance of an application window to get the program rolling.
		MainWindow window = newAppWindow(newData);
		
		//	Add any other files that are to be plotted.
		for (int i=1; i < theFiles.size(); ++i) {
			//	Read the file and add it to the current window.
            File theFile = theFiles.get(i);
            window.openFile(theFile);
		}
		window.setVisible(true);
		
		if (nonNull(newData)) {
			//	Read the plot definitions (if there are any).
			if (nonNull(defsFile))
				window.readPlotDefinitions(defsFile);
			
			//	Plot data if asked to.
			if (autoPlot) {
				selectAll(window.runList);
				window.handlePlot(null);
			}
		}
		
        return window;
	}
	
	/**
	*  Method that selects all the items in the specified list.
	*/
	private static void selectAll(JList list) {
		int size = list.getModel().getSize();
		list.setSelectionInterval(0,size-1);
	}
	
	/**
	*  Method that reads in a data file and return the resulting model data structure.
	*
	*  @param parent    The parent frame for dialogs (null is fine).
	*  @param theFile   The file to be read in.  If null is passed, this
	*                   method will do nothing.
	*  @param skipUnknown If <code>true</code>, then files with unknown types will be silently ignored.
	*                     If <code>false</code>, then files with unknown types are reported to the user.
	*/
	private static DataSet readData(Frame parent, File theFile, boolean skipUnkown) {
		ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
		DataSet data = null;
		
		if (nonNull(theFile)) {
			
			InputStream is = null;
			try {
				if (!theFile.exists())
					throw new IOException(
                            MessageFormat.format(resBundle.getString("fileDoesntExistErr"),
                            theFile.getName()));
				
                //  Is the file a GZIP file?
                boolean isGZIP = FileUtils.isGZIPCompressed(theFile);
                
				//	Create an input stream to the file.
				is = new FileInputStream(theFile);
                if (isGZIP) {
                    //  Convert the input stream into a GZIP input stream.
                    is = new GZIPInputStream(is);
                    
                    //  Convert the file name to remove any trailing ".gz".
                    String fileName = theFile.getName();
                    String ext = FileUtils.getExtension(fileName);
                    if ("gz".equals(ext)) {
                        int index = fileName.lastIndexOf('.');
                        fileName = fileName.substring(0, index);
                        theFile = new File(theFile.getParentFile(), fileName);
                    }
                    
                }
				
                //  Convert the plain input stream into a BufferedInputStream.
                BufferedInputStream bis = new BufferedInputStream(is);
                
				//	Get a reader for the specified file.
				DataReader reader = null;
				try {
					reader = DataReaderFactory.getReader(bis, theFile.getAbsolutePath());
				} catch (IOException e) {
					if (skipUnkown)	return null;	//	Silently cancel reading the file.
					throw e;						//	Re-throw the exception (which will be reported to the user below).
				}
				
				//	Read in the data (and keep only the 1st data set if there are more than one).
				List<DataSet> allData = reader.read(theFile.getPath(), bis);
				if (nonNull(allData))	data = allData.get(0);
				
			} catch( IOException ex ) {
                LOGGER.log(Level.SEVERE, null, ex);
                
				String msg = ex.getMessage();
				if (isNull(msg) && nonNull(ex.getCause()))
					msg = ex.getCause().getMessage();
				JOptionPane.showMessageDialog( parent, msg, resBundle.getString("ioErrorTitle"),
								JOptionPane.ERROR_MESSAGE );
				
			} catch ( IllegalArgumentException ex) {
                LOGGER.log(Level.SEVERE, null, ex);
				JOptionPane.showMessageDialog( parent, ex.getMessage(), resBundle.getString("errorTitle"),
								JOptionPane.ERROR_MESSAGE );
			
			} finally {
				//	Make sure and close the input stream, no matter what happens.
				try {
					if (is != null)	is.close();
				} catch (IOException ignore) {}
			}
		}
		
		return data;
	}
	
    /**
     * Creates, and initializes, a new application window with the specified content.
     * Effectively creates a new instance of this program.
     *
     * @param newData The data set this window contains. If null is passed, an
     *                  empty/default/new data set is created.
     * @return
     * @throws java.lang.NoSuchMethodException
     */
	public static MainWindow newAppWindow(DataSet newData)
							throws NoSuchMethodException {
	
		//	If the document data wasn't provided, create a default set of data.
		if (isNull(newData))	newData = DataSet.newInstance("Untitled");
		
		// Create an instance of this window.
		MainWindow appFrame = new MainWindow(newData);
		
		return appFrame;
	}
	
	
	/**
	*  Method called by the quit listener when the application has been asked to quit.
	*  Returns true if the quit should be canceled, or false if continuing with
	*  quitting the application is fine.
	*/
	private boolean handleQuit() {
		return false;
	}
	
    /**
     * Handle the user choosing "Open Plot Definition..." from the File menu. Lets the
     * user choose a data file and open it.
     *
     * @param event
     */
	public void handleOpenPlotDefs( ActionEvent event ) {
		
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
		String dir = MDIApplication.getInstance().getPreferences().getLastPath();
		ExtFilenameFilter filter = new ExtFilenameFilter("dat");
		filter.addExtension("txt");
		File theFile = AppUtilities.selectFile(this, FileDialog.LOAD, resBundle.getString("fileDialogLoad"),
										dir, null, filter);
		if (isNull(theFile))	return;
		
		//	Read and parse file.
		readPlotDefinitions(theFile);
	}


	/**
	*  Method that reads in the specified plot definitions file and stores it in this window.
	*  All exceptions are handled here.
	*
	*  @param theFile  Reference to the plot definitions file.
	*/
	private void readPlotDefinitions(File theFile) {
		if (isNull(theFile))	return;
		
        try (FileInputStream is = new FileInputStream(theFile)) {
            
			//	Read and parse file.
			_plotDefs = new PlotDefinitionList(is);
		
		} catch( IOException ex ) {
            LOGGER.log(Level.SEVERE, null, ex);
            
			String msg = ex.getMessage();
			if (isNull(msg) && nonNull(ex.getCause()))
				msg = ex.getCause().getMessage();
			
			JOptionPane.showMessageDialog( this, msg,
                    MDIApplication.getInstance().getResourceBundle().getString("ioErrorTitle"), 
                    JOptionPane.ERROR_MESSAGE );
		}
	    
	}
}
