
import java.util.List;
import java.util.AbstractList;
import java.util.ArrayList;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import javax.measure.quantity.Dimensionless;
import javax.measure.unit.Unit;

/**
 * A list of PlotDefinition objects used to store user defined plots.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: September 3, 2011
 * @version March 8, 2017
 */
public class PlotDefinitionList extends AbstractList<PlotDefinition> {

    //	The actual list of PlotDefinition objects stored.
    private final List<PlotDefinition> _defs = new ArrayList<>();

    /**
     * Create a new instance of a plot definition list using the specified input
     * stream to a plot definition data file.
     *
     * @param plotDefsInput An input stream containing the plot definition data.
     * @throws java.io.IOException If there is any problem reading from the input stream.
     */
    public PlotDefinitionList(InputStream plotDefsInput) throws IOException {

        //	Wrap the input stream in a line number reader.
        LineNumberReader reader = new LineNumberReader(new InputStreamReader(plotDefsInput));

        //	Start reading in the data
        String aLine;
        do {
            aLine = reader.readLine();
            //	Skip blank lines.
            if (aLine != null && !aLine.trim().equals("")) {

                //	Parse out the columns in this line.
                String[] parts = aLine.split("(,|\t)");
                if (parts.length != 5 && parts.length != 11)
                    throw new IOException("Expected either 5 or 11 plot definition parameters on line #" + reader.getLineNumber()
                            + " of the plot definition file, but got " + parts.length + ".");
                String indepName = parts[0].trim();
                String depName = parts[1].trim();
                String indepDesc = parts[2].trim();
                String depDesc = parts[3].trim();
                String title = parts[4].trim();

                Unit indepSrcUnits = Dimensionless.UNIT;
                Unit indepDestUnits = Dimensionless.UNIT;
                Unit depSrcUnits = Dimensionless.UNIT;
                Unit depDestUnits = Dimensionless.UNIT;
                if (parts.length > 5) {
                    try {
                        //  Skip the range parameter.
                        //  String range = parts[5].trim();

                        //  Get the independent source units.
                        String unitStr = parts[6].trim();
                        if (!"".equals(unitStr))
                            indepSrcUnits = Unit.valueOf(unitStr);

                        //  Get the independent destination units.
                        unitStr = parts[7].trim();
                        if (!"".equals(unitStr))
                            indepDestUnits = Unit.valueOf(unitStr);

                        if (!indepSrcUnits.isCompatible(indepDestUnits))
                            throw new IOException("Independent variable units are incompatible on line #"
                                    + reader.getLineNumber() + " of plot definition file.  Got "
                                    + indepSrcUnits + " and " + indepDestUnits);

                        //  Get the dependent source units.
                        unitStr = parts[8].trim();
                        if (!"".equals(unitStr))
                            depSrcUnits = Unit.valueOf(unitStr);

                        //  Get the dependent destination units.
                        unitStr = parts[9].trim();
                        if (!"".equals(unitStr))
                            depDestUnits = Unit.valueOf(unitStr);

                        if (!depSrcUnits.isCompatible(depDestUnits))
                            throw new IOException("Dependent variable units are incompatible on line #"
                                    + reader.getLineNumber() + " of plot definition file.  Got "
                                    + depSrcUnits + " and " + depDestUnits);
                        
                        //  Skip the dot type index.
                        //  String dotType = parts[10].trim();
                        
                    } catch (IllegalArgumentException e) {
                        throw new IOException("Illegal unit name found on line #"
                                + reader.getLineNumber() + " of the plot definition file.");
                    }
                }

                //	Store the values.
                PlotDefinition def = new PlotDefinition(title, indepName, indepDesc, depName, depDesc,
                        indepSrcUnits, indepDestUnits, depSrcUnits, depDestUnits);
                _defs.add(def);
            }
            
        } while (aLine != null);

        //	Reached the end of the file.
    }

    /**
     * Returns a string defining the parameters that must occur on each line of the plot
     * definition file.
     * 
     * @return A string defining the parameters that must occur on each line of the plot
     * definition file.
     */
    public static String plotDefinitionLine() {
        return "X-Variable, Y-Variable, X-Description, Y-Description, Plot Title, Plot-Range, X-Source Unit, X-Unit, Y-Source Unit, Y-Unit, Dots-Flag";
    }
    
    /**
     * The definition of each parameter in the plot definition file.
     * 
     * @return The definition of each parameter in the plot definition file.
     */
    public static List<String> plotDefinitionParameters() {
        List<String> list = new ArrayList();
        list.add("X-Variable = Name of the variable to plot on the X-axis");
        list.add("Y-Variable = Name of the variable to plot on the Y-axis");
        list.add("X-Description = A description or title for the variable plotted on the X-axis");
        list.add("Y-Description = A description or title for the variable plotted on the Y-axis");
        list.add("Plot Title = The title to give to the plot");
        list.add("Plot-Range = Not currently used, but must be present (put in a \":\")");
        list.add("X-Source Unit = The abbreviation used for the X-Variable source units (ignored if the data file has units defined)");
        list.add("X-Unit = The abbreviation used for the X-Variable units in the plot (must be consistent with X-Source Unit)");
        list.add("Y-Source Unit = The abbreviation used for the Y-Variable source units (ignored if the data file has units defined)");
        list.add("Y-Unit = The abbreviation used for the Y-Variable units in the plot (must be consistent with Y-Source Unit)");
        list.add("Dots-Flag = Not currently used, but must be present (put in a \"0\")");
        return list;
    }
    
    /**
     * Returns the plot definition at the specified position in this list.
     *
     * @param index index of element to return.
     * @return the plot definition at the specified position in this list.
     * @throws IndexOutOfBoundsException if the given index is out of range (index &lt; 0 ||
     *      index &ge; size()).
     */
    @Override
    public PlotDefinition get(int index) {
        return _defs.get(index);
    }

    /**
     * Return the number of plot definitions in this list.
     * 
     * @return The number of plot definitions in this list.
     */
    @Override
    public int size() {
        return _defs.size();
    }
}
