
import jahuwaldt.io.ExtFilenameFilter;
import jahuwaldt.js.datareader.*;
import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.MDIApplication;
import jahuwaldt.swing.SortedListModel;
import jahuwaldt.swing.SortedListModel.SortOrder;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;
import java.util.ResourceBundle;
import javax.measure.quantity.Dimensionless;
import javax.measure.unit.Unit;
import javax.swing.*;
import javax.swing.border.Border;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javolution.util.FastTable;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.labels.StandardXYToolTipGenerator;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.chart.title.LegendTitle;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

/**
 * An application window that displays a quick plot of the data contained in a DataSet
 * made up of DataCase objects containing the same set of ArrayParam objects. This allows
 * "overlay" plots of identical parameters from different "runs" or cases.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: September 3, 2011
 * @version December 22, 2023
 */
public class PlotWindow extends AbstractPlotWindow {

    private static final long serialVersionUID = 1L;

    /**
    *  The panel that contains the plot region of our window.  Sub-classes may
    *  use this to access the ChartPanel it contains in it's CENTER and the
    *  button panel it contains in it's SOUTH.
    */
    protected JPanel _centerPanel = new JPanel(new BorderLayout());
    
    //  The index in the available parameter list of the element that the mouse is current over.
    private int _mouseOverIndex = 0;
    
    //  The check-box for showing or hiding the legend.
    private JCheckBox _showLegendCB = null;
    
    //  The legend for the plot is stored here.
    private LegendTitle _plotLegend = null;

    // A reference to the all the available data.
    private DataSet _allData;

    // A reference to the particular arrays we are doing the plot of.
    private final List<ArrayParam> _plottedData = new ArrayList<>();

    //  The independent parameters for each of the plotted arrays in _plottedData.
    private final List<ArrayParam> _indepData = new ArrayList<>();

    //  A list of the parameters for plotting.
    private JList _paramList;
    
    //  A parameter list model that allows sorting of the parameters.
    private SortedListModel _sortedListModel;
    
    //  Flag used to indicate that a JList selection change is due to a change in the sort order.
    private boolean _autoSelectionChange = false;
    

    /**
    *  Creates a new instance of a data plot window using the
    *  specified list of data arrays.
    *
    *  @param  title  Title of the window.
    *  @param  data   A DataCase containing a single case to be plotted.
     * @throws java.lang.NoSuchMethodException
    *  @throws IllegalArgumentException if the input data set can not be plotted.
    */
    public PlotWindow( String title, DataCase data ) throws NoSuchMethodException, IllegalArgumentException {
        super( title, title);
        
        DataSet tmpSet = DataSet.newInstance("tmp");
        tmpSet.add(data);
        initialize(tmpSet);
        
    }
    
    /**
    *  Creates a new instance of a data plot window using the
    *  specified list of data arrays.
    *
    *  @param  title  Title of the window.
    *  @param  data   DataSet containing the cases to be plotted (all cases must have the
    *                 same set of parameters).
     * @throws java.lang.NoSuchMethodException
    *  @throws IllegalArgumentException if the input data set can not be plotted.
    */
    public PlotWindow( String title, DataSet data )
                        throws NoSuchMethodException, IllegalArgumentException {
        super( title, title);
        
        initialize(data);
    }
    
    /**
    *  Used to initialize this plot window.
    */
    private void initialize(DataSet data) throws NoSuchMethodException, IllegalArgumentException {
        
        //  Make sure the input data set is compatible with this plot window.
        int compatibleFlag = checkForMultiCaseCompatibility(data);
        if (compatibleFlag == 1)
            throw new IllegalArgumentException(MDIApplication.getInstance().getResourceBundle().getString("inconsistantArrayLengthsErr"));
        else if (compatibleFlag == 2)
            throw new IllegalArgumentException(MDIApplication.getInstance().getResourceBundle().getString("inconsistantArrayNamesErr"));
        
        //  Store the initially plotted parameters from each case.
        for (DataCase aCase : data) {
            List<ArrayParam> caseArrays = aCase.getAllArrays();
            _indepData.add(caseArrays.get(0));
            _plottedData.add(caseArrays.get(1));
            FastTable.recycle((FastTable<ArrayParam>)caseArrays);
        }       
        
        // Set up some class variables.
        _allData = data;
        
        //  Layout the contents of this window.
        layoutWindow();
    }
    
    
    /**
    *  Checks to make sure all the cases and arrays in the specified data set
    *  are compatible with a multi-case plotting window.  To be compatible, each case must have the same number
    *  of arrays and arrays at the same index in each case must have the same names.  If the
    *  cases are compatible, this method will return 0.
    *
    *  @param data  The DataSet to check for compatibility with this plot window.
    *  @return A value of 0 if the cases are compatible with plotting in this window. A value of
    *          1 if the cases have different numbers of arrays.  A value of 2 if the arrays in
    *          the cases do not have the same names.
    */
    public static int checkForMultiCaseCompatibility(DataSet data) {
        
        List<ArrayParam> refArrays = data.get(0).getAllArrays();
        
        for (DataCase aCase : data) {
            
            //  Check that each case as the same number of arrays.
            int size = refArrays.size();
            List<ArrayParam> caseArrays = aCase.getAllArrays();
            if (caseArrays.size() != size)
                return 1;
            
            //  Check that each array has the same name.
            for (int i=0; i < size; ++i) {
                CharSequence refName = refArrays.get(i).getName();
                if (!caseArrays.get(i).getName().equals(refName))
                    return 2;
            }
            
            FastTable.recycle((FastTable<ArrayParam>)caseArrays);
        }
        FastTable.recycle((FastTable<ArrayParam>)refArrays);
        
        return 0;
    }
    
    /**
    *  Method that lays out the contents of this window.
    */
    private void layoutWindow() throws NoSuchMethodException {
    
        // Set up this window's user interface.
        Container cp = getContentPane();
        
        //  Create a split pane with the parameters to plot on the left and the plot itself on the right.
        JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
        splitPane.setRightComponent(_centerPanel);
        cp.add(splitPane, BorderLayout.CENTER);
        
        
        //  Place a list of available plotting parameters on the left side of the screen.
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
        List<ArrayParam> allArrays = _allData.get(0).getAllArrays();
        if (allArrays.size() > 2) {
            Box vbox = Box.createVerticalBox();
            Border outsideBorder = BorderFactory.createEmptyBorder(4,10,10,0);
            Border insideBorder = BorderFactory.createTitledBorder(resBundle.getString("plottableParametersText"));
            vbox.setBorder(BorderFactory.createCompoundBorder(outsideBorder, insideBorder));
            splitPane.setLeftComponent(vbox);
            
            //  Create a list for displaying the parameter names.
            DefaultListModel unsortedListModel = createDefaultListModel(_allData.get(0));
            _sortedListModel = new SortedListModel(unsortedListModel, SortOrder.UNORDERED);
            _paramList = new JList(_sortedListModel);
            int selMode = (_allData.size() > 1 ?
                            ListSelectionModel.SINGLE_SELECTION : ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
            _paramList.setSelectionMode(selMode);
            _paramList.setLayoutOrientation(JList.VERTICAL);
            _paramList.setPrototypeCellValue("A Longish Name");
            _paramList.setSelectedIndex(1);
            _paramList.addListSelectionListener(new ListSelectionListener() {
                @Override
                public void valueChanged(ListSelectionEvent e) {
                    if (e.getValueIsAdjusting() == false && !_autoSelectionChange) {
                        JList list = (JList)e.getSource();
                        int idx = list.getSelectedIndex();
                        if (idx >= 0) {
                            //  Convert the selected (sorted) indexes to unsorted indexes.
                            int[] unsortedIndices = sorted2Unsorted(_sortedListModel, list.getSelectedIndices());
                            Unit[] units = new Unit[unsortedIndices.length];
                            
                            //  Change plot to the selected parameters.
                            setDependents(unsortedIndices, units);
                            notifySelectionChanged();
                            regeneratePlot();
                        }
                    }
                }
            });
            
            //  Track where the mouse is.
            _paramList.addMouseMotionListener(new MouseMotionAdapter() {
                @Override
                public void mouseMoved(MouseEvent e) {
                    JList list = (JList)e.getComponent();
                    _mouseOverIndex = list.locationToIndex(e.getPoint());
                }
            });
            
            //  Modifiy the cell renderer to use an appropriate tool tip text string for each element.
            _paramList.setCellRenderer(new DefaultListCellRenderer() {
                private static final long serialVersionUID = 1L;
                int lastIdx = -1;
                String lastString = null;
                @Override
                public String getToolTipText(MouseEvent event) {
                    if (_mouseOverIndex != lastIdx) {
                        lastIdx = _mouseOverIndex;
                        List<ArrayParam> allArrays = _allData.get(0).getAllArrays();
                        ArrayParam param = allArrays.get(lastIdx);
                        Object userObject = param.getUserObject();
                        if (nonNull(userObject))
                            lastString = userObject.toString();
                        else
                            lastString = null;
                    }
                    return lastString;
                }
            });
            
            JScrollPane listScroller = new JScrollPane(_paramList);
            listScroller.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
            listScroller.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
            vbox.add(listScroller);
            
            
            //  Add sort order buttons.
            ButtonGroup bgSortOrder = new ButtonGroup();
            Box panelSort = Box.createVerticalBox();
            panelSort.setBorder(BorderFactory.createTitledBorder(resBundle.getString("sortOrderTitle")));
            
            //  Sort Ascending button.
            JRadioButton rbAscendingSort = new JRadioButton(resBundle.getString("sortOrderAscending"));
            bgSortOrder.add(rbAscendingSort);
            rbAscendingSort.setSelected(false);
            rbAscendingSort.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
            rbAscendingSort.setMargin(new java.awt.Insets(0, 0, 0, 0));
            rbAscendingSort.addActionListener(new SortOrderRBActionListener(SortOrder.ASCENDING));
            panelSort.add(rbAscendingSort);
            
            JRadioButton rbDescendingSort = new JRadioButton(resBundle.getString("sortOrderDescending"));
            bgSortOrder.add(rbDescendingSort);
            rbDescendingSort.setSelected(false);
            rbDescendingSort.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
            rbDescendingSort.setMargin(new java.awt.Insets(0, 0, 0, 0));
            rbDescendingSort.addActionListener(new SortOrderRBActionListener(SortOrder.DESCENDING));
            panelSort.add(rbDescendingSort);
            
            JRadioButton rbUnordered = new JRadioButton(resBundle.getString("sortOrderUnordered"));
            bgSortOrder.add(rbUnordered);
            rbUnordered.setSelected(true);
            rbUnordered.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
            rbUnordered.setMargin(new java.awt.Insets(0, 0, 0, 0));
            rbUnordered.addActionListener(new SortOrderRBActionListener(SortOrder.UNORDERED));
            panelSort.add(rbUnordered);
            
            vbox.add(panelSort);
        }
        
        
        //  Create user interface buttons.
        JButton changeIndepButton = new JButton( resBundle.getString("changeIndepBtnText") );
        changeIndepButton.setHorizontalAlignment(SwingConstants.CENTER);
        changeIndepButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent evt) {
                doChangeIndep();
            }
        } );
        changeIndepButton.setEnabled(allArrays.size() > 2);
        changeIndepButton.setToolTipText( resBundle.getString("changeIndepBtnToolTip") );

        JButton closeButton = new JButton( resBundle.getString("closeItemText") );
        this.getRootPane().setDefaultButton(closeButton);
        closeButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent evt) {
                handleClose(null);
            }
        } );
        closeButton.setToolTipText( resBundle.getString("closeItemToolTip") );
        
        //  Create a button panel and add all the buttons to it.
        Box buttonPanel = Box.createHorizontalBox();
        buttonPanel.setAlignmentX(Box.LEFT_ALIGNMENT);
        buttonPanel.add(changeIndepButton);
        buttonPanel.add(Box.createGlue());
        buttonPanel.add(closeButton);
        buttonPanel.add(Box.createHorizontalStrut(20));
        buttonPanel.setBorder(BorderFactory.createEmptyBorder(10,10,10,10));
        
        //  Create a panel of options.
        Box optionPanel = Box.createHorizontalBox();
        optionPanel.setAlignmentX(Box.LEFT_ALIGNMENT);
        _showLegendCB = new JCheckBox(resBundle.getString("showLegendText"), false);
        _showLegendCB.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent evt) {
                JCheckBox cb = (JCheckBox)evt.getSource();
                setLegendVisible(cb.isSelected());
            }
        });
        optionPanel.add(_showLegendCB);
        
        //  Add the options and buttons to the bottom of the window.
        Box vbox = Box.createVerticalBox();
        vbox.add(optionPanel);
        vbox.add(buttonPanel);
        _centerPanel.add(vbox, BorderLayout.SOUTH);
        
        //  Create the plot panel containing the plot itself.
        ChartPanel plotPanel = createPlotPanel(_plottedData, _indepData, _allData, this);
        _centerPanel.add(plotPanel, BorderLayout.CENTER);
        
        //  Set the visibility of the legend.
        _plotLegend = plotPanel.getChart().getLegend();
        setLegendVisible(_showLegendCB.isSelected());
        
    }
    
    /**
    *  Method called whenever the user changes the selection of plotted parameters.
    *  The default implementation does nothing.
    */
    protected void notifySelectionChanged() { }
    
    /**
    *  Creates a default list model containing the specified array of elements.
    */
    private static DefaultListModel createDefaultListModel(DataCase data) {
        DefaultListModel model = new DefaultListModel();
        
        List<ArrayParam> allArrays = data.getAllArrays();
        int size = allArrays.size();
        for (int i=0; i < size; ++i) {
            ArrayParam element = allArrays.get(i);
            StringBuilder name = new StringBuilder(element.getName().toString());
            if (name.toString().equals("")) {
                name.append("Param");
                name.append(i+1);
            }
            Unit unit = element.getUnit();
            if ( !Dimensionless.UNIT.equals(unit) ) {
                name.append(" (");
                name.append(unit);
                name.append(")");
            }
            model.addElement(name.toString());
        }
        
        return model;
    }
    
    /**
    *  Return a list of unsorted JList indexes from the supplied list of sorted JList indexes.
    */
    private static int[] sorted2Unsorted(SortedListModel sortedListModel, int[] sortedIndexes) {
        int size = sortedIndexes.length;
        int[] unsortedIndexes = new int[size];
        for (int i=0; i < size; ++i)
            unsortedIndexes[i] = sortedListModel.toUnsortedModelIndex(sortedIndexes[i]);
        return unsortedIndexes;
    }
    
    /**
    *  Return a list of sorted JList indexes from the supplied list of sorted JList indexes.
    */
    private static int[] unsorted2Sorted(SortedListModel sortedListModel, int[] unsortedIndexes) {
        int size = unsortedIndexes.length;
        int[] sortedIndexes = new int[size];
        for (int i=0; i < size; ++i)
            sortedIndexes[i] = sortedListModel.toSortedModelIndex(unsortedIndexes[i]);
        return sortedIndexes;
    }
    
    /**
    *  An action listener for the list sort order radio button group.
    */
    private class SortOrderRBActionListener implements ActionListener {
        private final SortOrder sortOrder;
        
        /**
        *  Construct a sort order button action listener that uses
        *  the specified sort order.
        */
        SortOrderRBActionListener(SortOrder order) {
            sortOrder = order;
        }
        
        @Override
        public void actionPerformed(ActionEvent e) {
            //  Get the list of currently selected indicies.
            int[] sortedIndices = _paramList.getSelectedIndices();
            if (nonNull(sortedIndices) && sortedIndices.length > 0) {
            
                //  Convert the selected (sorted) indexes to unsorted indexes.
                int[] unsortedIndices = sorted2Unsorted(_sortedListModel, sortedIndices);
                
                //  Change the sort order.
                _sortedListModel.setSortOrder(sortOrder);
                
                //  Reselect the originally selected items (now sorted into the list differently).
                sortedIndices = unsorted2Sorted(_sortedListModel, unsortedIndices);
                _autoSelectionChange = true;
                _paramList.setSelectedIndices(sortedIndices);
                _paramList.ensureIndexIsVisible(sortedIndices[0]);
                _autoSelectionChange = false;
            }
        }
    }
    
    
    /**
    *  Method that will build an "Export" menu containing a list of all the DataReader objects that
    *  are capable of writing to files.
    *
    *  @param  title    The title for the menu that is created.
     * @return The newly created Export menu.
    *  @returns A JMenu instance containing a list of all DataReader objects that can write to
    *           files in various formats.  If there are no appropriate readers available,
    *           null is returned.
    */
    @Override
    protected JMenu buildExportMenu(String title) {
    
        //  Get the list of data readers that are available.
        DataReader[] allReaders = DataReaderFactory.getAllReaders();
        if (isNull(allReaders) || allReaders.length < 1)
            return null;

        //  Create the menu
        JMenu menu = new JMenu(title);
        
        //  Loop over all the readers.
        boolean noneFound = true;
        int numReaders = allReaders.length;
        for (int i=0; i < numReaders; ++i) {
            DataReader reader = allReaders[i];
            
            if (reader.canWriteData()) {
                //  Found a "writer".
                noneFound = false;
                
                //  Create a menu item for it.
                JMenuItem menuItem = new JMenuItem(reader.toString() + "...");
                menuItem.addActionListener( new ExportAsMenuListener(reader) );
                menuItem.setActionCommand( reader.toString() );
                menuItem.setEnabled( true );
                
                //  Add the menu item to the menu.
                menu.add(menuItem);
            }
        }
        
        if (noneFound)  menu = null;
        
        return menu;
    }


    /**
    *  Defines an action listener for our Export menu items.
    */
    private class ExportAsMenuListener implements ActionListener {
        //  Keep a reference to the reader used by this menu.
        private final DataReader reader;
        
        public ExportAsMenuListener(DataReader reader) {
            this.reader = reader;
        }
        
        @Override
        public void actionPerformed( ActionEvent evt ) {
                handleSaveAs();
        }

        /**
        *  Handles saving the data using this menu item's DataReader instance.
        */
        private void handleSaveAs() {

            //  Ask the user if they want to export all the data or only selected runs.
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
            Object[] options = {resBundle.getString("exportPlotOptionAll"),
                                resBundle.getString("exportPlotOptionSel"),
                                resBundle.getString("cancel")};
            String msg = resBundle.getString("exportPlotMsg");
            String title = resBundle.getString("exportPlotTitle");
            int choice = JOptionPane.showOptionDialog(PlotWindow.this, msg, title,
                    JOptionPane.YES_NO_CANCEL_OPTION,JOptionPane.QUESTION_MESSAGE,
                    null, options, options[0]);
            if (choice == JOptionPane.CLOSED_OPTION || choice == JOptionPane.CANCEL_OPTION) return;
            boolean exportAllData = (choice == JOptionPane.YES_OPTION);
            
            //  Set up the file name filter.
            String extension = reader.getExtension();
            ExtFilenameFilter filter = new ExtFilenameFilter(extension, reader.toString() + " File");

            //  Build the directory and filename to prompt the user with.
            String dir = MDIApplication.getInstance().getPreferences().getLastPath();
            String fileName = getDataName();
            
            //  Replace any ":" characters with "_".
            fileName = fileName.replace(':', '_');
            
            //  Replace any "_ " with " ".
            fileName = fileName.replaceAll("_ ", " ");
            
            int idx = fileName.lastIndexOf(".");
            if (idx > 0)
                fileName = fileName.substring(0, idx);
            fileName += "." + extension;
            
            //  Ask the user to select a file for saving.
            File chosenFile = AppUtilities.selectFile4Save(PlotWindow.this,
                    MessageFormat.format(resBundle.getString("fileSaveDialogMsg"),reader.toString()),
                    dir, fileName, filter, extension,
                    resBundle.getString("fileExistsMsgFmt"), resBundle.getString("warningTitle"));
            if (nonNull(chosenFile)) {
                if (chosenFile.exists() && !chosenFile.canWrite()) {
                    JOptionPane.showMessageDialog( PlotWindow.this,
                                resBundle.getString("canNotWrite2File"),
                                resBundle.getString("ioErrorTitle"), JOptionPane.ERROR_MESSAGE );
                    return;
                }

                //  Export the data.
                exportData(reader, chosenFile, exportAllData);
                
                //  Store off the path reference for later use.
                MDIApplication.getInstance().getPreferences().setLastPath(chosenFile.getParent());
            }
        }


        /**
        *  Method that exports the specified data to the specified file using
        *  the specified writer.
        *
        *  @param writer     The DataReader to use to write out the data.
        *  @param outFile    Reference to the file path where the file should be written.
        *  @param exportAllData Set to true to export all the data and false to output only selected data.
        */
        private void exportData(DataReader writer, File outFile, boolean exportAllData) {

            //  Create a List of DataSet objects as required by the writer.
            List<DataSet> list = new ArrayList<>();
            DataSet set = DataSet.newInstance("Temporary Set");

            //  Add the selected data to the list of data to be output.
            if (exportAllData) {
                for (DataCase aCase : _allData) {
                    DataCase selectedData = DataCase.newInstance("Temporary Case");
                    selectedData.addAll(aCase.getAllArrays());
                    set.add(selectedData);
                }

            } else {
                DataCase selectedData = DataCase.newInstance("Temporary Case");
                set.add(selectedData);

                int depPos = 0;
                int numSets = _indepData.size();    //  The indepData size indicates how many sets of data are plotted.
                int numArrays = _plottedData.size() / numSets;
                for (int setNum = 0; setNum < numSets; ++setNum) {

                    //  Add the independent parameter.
                    selectedData.add(_indepData.get(setNum));

                    //  Loop over the number of variables to be plotted in each set.
                    for (int i = 0; i < numArrays; ++i) {
                        //  Add the dependent parameter.
                        selectedData.add(_plottedData.get(depPos));
                        ++depPos;
                    }
                }

            }
            list.add(set);

            try (FileOutputStream output = new FileOutputStream(outFile)) {
                
                // Write out the selected tables using the specified writer.
                writer.write( output, list );
                
            } catch( IOException ex ) {
                outFile.delete();
                ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
                JOptionPane.showMessageDialog( PlotWindow.this, ex.getMessage(), resBundle.getString("ioErrorTitle"),
                            JOptionPane.ERROR_MESSAGE );
            }
        }
        
    }   //  end class ExportAsMenuListener


    /**
    *  Creates a panel that contains the plot of the data.
    *
    *  @param  plottedData  List of dependent ArrayParam objects to be plotted.
    *  @param  indepData    List of independent ArrayParam objects for each dependent array.
    *  @param  allData     The DataSet containing all the available data.
    *  @param  target      The object that receives the action events.  Must have the
    *                      following methods defined:  "handlePrint(ActionEvent e)", and
    *                      "handlePageSetup(ActionEvent e)".
    */
    private ChartPanel createPlotPanel(List<ArrayParam> depData, List<ArrayParam> indepData, DataSet allData, Object target)
                            throws NoSuchMethodException {
    
        //  Create the chart.
        JFreeChart chart = createChart(depData, indepData, allData);
        
        //  Create a panel to put the chart in.
        createChartPanel(chart, target);
        
        //  Return a reference to the newly created chart panel.
        return getChartPanel();
    }
    

    /**
    *  Organizes the input data into series for plotting and creates
    *  an XY chart object.
    */
    private static JFreeChart createChart(List<ArrayParam> depData, List<ArrayParam> indepData, DataSet allData) {

        //  Build up a run list.
        XYSeriesCollection runList = new XYSeriesCollection();
        
        List<String> depNames = new ArrayList<>();
        int depPos = 0;
        int numSets = indepData.size(); //  The indepData indicates how many sets of data are plotted.
        int numArrays = depData.size()/numSets;
        for (int setNum=0; setNum < numSets; ++setNum) {
            ArrayParam indepArray = indepData.get(setNum);
            CharSequence dataSetName = allData.get(setNum).getName();
            
            //  Loop over the number of variables to be plotted in each set.
            for (int i=0; i < numArrays; ++i) {
                ArrayParam depArray = depData.get(depPos);
                if (numSets == 1) {
                    dataSetName = depData.get(depPos).getName();
                    depNames.add(dataSetName.toString());
                }
                runList.addSeries(createRun(indepArray, depArray, dataSetName));
                ++depPos;
            }
        }
        
        //  Range axis label
        String yAxisLabel;
        if (depNames.size() < 2)
            //  There is only a single dependent parameter plotted for, potentially, multiple data sets.
            yAxisLabel = createParamLabel(depData.get(0));
        
        else {
            //  There are, potentially, multiple parameters plotted for a single data set.
            
            //  Build up a label made up of the names of all the plotted paramters.
            StringBuilder buffer = new StringBuilder();
            int size = depNames.size();
            for (int i=0; i < size; ++i) {
                buffer.append(depNames.get(i));
                if (i != size - 1)
                    buffer.append(", ");
            }
            yAxisLabel = buffer.toString();
        }
        
        //  Domain axis label.
        String xAxisLabel = createParamLabel(indepData.get(0));
        
        //  Create the chart.
        JFreeChart chart = ChartFactory.createXYLineChart(
                                                null,                       //  No Title
                                                xAxisLabel,                 //  Domain axis label
                                                yAxisLabel,                 //  Range axis label
                                                runList,                    //  Data to be plotted
                                                PlotOrientation.VERTICAL,   //  Vertical orientation
                                                true,                       //  Include legend
                                                true,                       //  Include tooltips
                                                false);                     //  Do not include URLs
        
        // set the background color for the chart...
        chart.setBackgroundPaint(Color.WHITE);
        
        //  Get a reference to the plot.
        XYPlot plot = (XYPlot)chart.getPlot();

        //  Create a renderer that shows lines, no plot symbols, and has tool-tip values shown.
        XYLineAndShapeRenderer renderer = new XYLineAndShapeRenderer(true,false);
        renderer.setBaseToolTipGenerator(new StandardXYToolTipGenerator());
        
        //  If there are less than 100 data points, use plot symbols.
        if (indepData.get(0).size() < 100)
            renderer.setBaseShapesVisible(true);

        //  Change the color for the line drawn from red to blue.
        renderer.setSeriesStroke(0, new java.awt.BasicStroke(1.2F));
        
        //  Make the plot use the new renderer.
        plot.setRenderer(renderer);
        
        return chart;
    }
    
    /**
    *  Creates a single series or run for plotting from the specified dependent & independent arrays.
    */
    private static XYSeries createRun(ArrayParam indep, ArrayParam dependent, CharSequence runName) {

        XYSeries run = new XYSeries(runName.toString(), false);
        int length = indep.size();
        for (int i=0; i < length; ++i) {
            double indValue = indep.getValue(i);
            double depValue = dependent.getValue(i);
            run.add(indValue, depValue);
        }

        return run;
    }

    /**
    *  Method that creates a string representation of a parameter in the following format:
    *  "[refData], name, [units]".  If the reference data is null, it is not included.  If
    *  the units are Dimensionless, they are not included.
    */
    private static String createParamLabel(UnitParameter param) {
        StringBuilder buffer = new StringBuilder();
        
        Object refData = param.getUserObject();
        if (nonNull(refData)) {
            buffer.append(refData);
            buffer.append(", ");
        }
        buffer.append(param.getName());
        Unit unit = param.getUnit();
        if ( !Dimensionless.UNIT.equals(unit) ) {
            buffer.append(" (");
            buffer.append(unit);
            buffer.append(")");
        }
        
        return buffer.toString();
    }
    
    /**
    *  Method that shows or hides the legend.
    */
    private void setLegendVisible(boolean visible) {
        JFreeChart chart = getChartPanel().getChart();
        if (visible) {
            chart.removeLegend();
            chart.addLegend(_plotLegend);
        } else {
            _plotLegend = chart.getLegend();
            chart.removeLegend();
        }
    }
    

    /**
    *  Method that regenerates and redraws the plot in the window.  The old
    *  plot is disposed of and a new plot and plot panel is created.
    */
    protected void regeneratePlot() {
        try {
        
            //  Remove the old plot.
            if (getChartPanel() != null)
                _centerPanel.remove(getChartPanel());
            
            //  Create the new plot and add it to the window.
            ChartPanel plotPanel = createPlotPanel(_plottedData, _indepData, _allData, this);
            _centerPanel.add(plotPanel, BorderLayout.CENTER);
            
            //  Set the visibility of the legend.
            _plotLegend = plotPanel.getChart().getLegend();
            setLegendVisible(_showLegendCB.isSelected());
            
            //  Have the window re-configure itself.
            validate();
            
        } catch( NoSuchMethodException e ) {
            e.printStackTrace();
        }

    }
    

    /**
    *  Method that handles the user clicking on the "Change Independent" button.
    *  Asks the user to select a new independent variable for plotting.
    */
    private void doChangeIndep() {
        DataCase aCase = DataCase.newInstance("Temporary Case");
        aCase.addAll(_allData.get(0).getAllArrays());
        DataSet input = DataSet.newInstance("Temporary Set");
        input.add(aCase);
        
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
        SelectParamsDialog dialog = new SelectParamsDialog(this, resBundle.getString("selParameterText1"),
                    resBundle.getString("selParameterText2"), input, true, false);
        dialog.setLocation(AppUtilities.dialogPosition(dialog));
        dialog.setVisible(true);
        
        DataSet selection = dialog.getSelected();
        if (nonNull(selection)) {
            //  Get the selected array name.
            aCase = selection.get(0);
            CharSequence indepName = aCase.get(0).getName();
            
            //  Get the array with the selected name from each case.
            _indepData.clear();
            for ( DataCase nextCase : _allData ) {
                ArrayParam indep = (ArrayParam)nextCase.get(indepName);
                _indepData.add(indep);
            }
            
            regeneratePlot();
        }
        
        notifySelectionChanged();
    }

    /**
     * Sets the independent variable used for plotting to the specified array. The plot
     * display is not redrawn by this method.
     *
     * @param indepIndex Index into the parameter database of the new independent variable.
     * @param unit The unit in which the data is to be plotted.  May pass null which
     * means "no unit conversions".
     * @see #regeneratePlot
     */
    protected void setIndependent(int indepIndex, Unit unit) {
        _indepData.clear();
        for (DataCase aCase : _allData) {
            if (isNull(unit) || Dimensionless.UNIT.equals(unit))
                _indepData.add(aCase.getAllArrays().get(indepIndex));
                
            else
                _indepData.add(aCase.getAllArrays().get(indepIndex).to(unit));
        }
    }

    /**
     * Sets the dependent parameters being plotted to those with the specified indexes.
     * The plot display is not redrawn by this method.
     *
     * @param idxArr A list of indexes into the parameter database of the new dependent
     * variables.
     * @param unitArr A list of units that go with each parameter referenced by idxArr.
     *                  May contain a list of null values meaning "no unit conversions".
     * @see #regeneratePlot
     */
    protected void setDependents(int[] idxArr, Unit[] unitArr) {
        _plottedData.clear();
        for (DataCase aCase : _allData) {
            List<ArrayParam> allArrays = aCase.getAllArrays();
            int length = idxArr.length;
            for (int i = 0; i < length; ++i) {
                if (isNull(unitArr[i]) || Dimensionless.UNIT.equals(unitArr[i]))
                    _plottedData.add(allArrays.get(idxArr[i]));
                else
                    _plottedData.add(allArrays.get(idxArr[i]).to(unitArr[i]));
            }
        }

        //  Change the selections in the GUI list.
        _autoSelectionChange = true;
        _paramList.setSelectedIndices(unsorted2Sorted(_sortedListModel, idxArr));
        _autoSelectionChange = false;
    }
}

