/*
 *   POST2DataReader  -- Reads data from a POST II PROFIL binary data file.
 *
 *   Copyright (C) 2011-2014 by Joseph A. Huwaldt
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.datareader;

import jahuwaldt.swing.AppUtilities;
import java.awt.Frame;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.DoubleBuffer;
import java.util.List;
import javax.measure.quantity.Dimensionless;
import javolution.context.ArrayFactory;
import javolution.text.Text;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Integer64;

/**
 * This class translates data between a list of DataSet data structures and a
 * stream containing POST II PROFIL data.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: September 3, 2011
 * @version April 27, 2014
 */
public class POST2DataReader implements DataReader {

	private static final boolean DEBUG = false;
	
	//	The default name given to data sets.
	private CharSequence _defaultSetName = "POST II Data";

	//  A brief description of the data read by this reader.
	private static final String DESCRIPTION = "POST II PROFIL Data";
	
	//	The size, in bytes, of a record block in the PROFIL file.
	private static final int BLOCK_SIZE = 13312;
	
	/**
	*  Returns a string representation of the object.  This will return a brief description of the format
	*  read by this reader.
	*/
    @Override
	public String toString() {
		return DESCRIPTION;
	}
	
	/**
	*  Returns the preferred file extension (not including the ".") for files of
	*  this DataReader's type.
	*/
    @Override
	public String getExtension() {
		return "prf";
	}
	
	/**
	* Compares this object with the specified object for order based on the
	* <code>toString().compareTo(o.toString())</code> method. Returns a negative integer,
	* zero, or a positive integer as this object is less than, equal to, or
	* greater than the specified object.
	*/
    @Override
	public int compareTo(DataReader o) {
		return this.toString().compareTo(o.toString());
	}
	
	/**
	*  Method that determines if this reader can read data from the specified input stream.
	*
	*  @param  pathName The path to the file to be read.
	*  @param  input  An input stream containing the data to be read.  Any methods that
	*                 read from this stream must first set a mark and then reset back to
	*                 that mark before the method returns (even if it returns with an
	*                 exception).
	*  @returns DataReader.NO if the file is not recognized or DataReader.YES if the
	*           appears to be readable.
	*/
    @Override
	public int canReadData(String pathName, BufferedInputStream input) throws IOException {

		int response = YES;
		try {
			//	Check the format of the file.  If an exception is thrown, then
			//	we can't read the file.
			try {
				//	Mark the current place in the input stream so we can return to it when we are done.
				input.mark(BLOCK_SIZE+8);	//	We will be reading in a 13312 byte header.
			
				//	Read in the header data, checking for validity.
				readHeaderData(input);

			} finally {
				//	Reset the buffer back to where it was before reading from it.
				input.reset();
			}
			
		} catch (Exception e) {
			response = NO;
		}
		
		return response;
	}
	
	/**
	*  Returns false.  This class can not write data to a MatLab V4 formatted file stream.
	*/
    @Override
	public boolean canWriteData() {
		return false;
	}
	
	/**
	*  The POST2 PROFIL format can save only a single case of arrays.
	*
	*  @param parent  Determines the Frame in  which the dialog is displayed; if null,
	*                 or if the parentComponent has no  Frame, a default Frame is used.
	*  @param data    The input data set that is to be selected for saving.
	*  @return  A list of DataSet objects containing the selected data to be saved.
	*           Could return null if the user selects nothing.
	*/
    @Override
	public List<DataSet> selectDataForSaving(Frame parent, List<DataSet> data) {
	
		if (data == null || data.size() < 1)	return null;

		DataSet selected;
		
		if (data.size() == 1 && data.get(0).size() == 1)
			//  If there is just one case, save that.
			selected = data.get(0);
		
		else {
			//  If there is more than one case, 
			//	ask the user to select cases to export.
			SelectCasesDialog dialog = new SelectCasesDialog(parent, "Select a case to Export",
												"Select a single case from the list below:", data, true, true);
			dialog.setLocation(AppUtilities.dialogPosition(dialog));
			dialog.setVisible(true);

			//	Retrieve the selected case.
			selected = (DataSet)dialog.getSelected();
			dialog.dispose();
		}
		if (selected == null || selected.size() < 1) return null;

		FastTable<DataSet> output = FastTable.newInstance();
		output.add(selected);

		return output;
	}
	
	
	/**
	*  Method that sets the default set name to use for
	*  MatLab V4 array data sets.
	*/
    @Override
	public void setDefaultSetName(CharSequence name) {
		_defaultSetName = name;
	}

	
	/**
	*  Method that reads in POST II PROFIL file data
	*  from the specified input stream and returns that data as a list
	*  of {@link DataSet} objects (containing a single {@link DataCase} containing an
	*  {@link ArrayParam} for each parameter in the PROFIL file).
	*
	*  @param  pathName The path to the file to be read.
	*  @param   input  An input stream containing the POST II PROFIL data.
	*  @returns A list of {@link DataSet} objects that contains the data read
	*           in from the specified stream (will contain a single data set
	*           which will contain a single case containing ArrayParams for
	*           each parameter in the PROFIL file).
	*  @throws IOException If there is a problem reading the specified stream.
	*/
    @Override
	public List<DataSet> read(String pathName, InputStream input) throws IOException {

		//	Create an empty list to store DataSets into.
		FastTable<DataSet> dataSets = FastTable.newInstance();

		//	Read in the header data.
		HeaderData hd = readHeaderData(input);
		
		//	Read in the parameter names.
		FastTable<Text> paramNames = readParamNames(hd, input);
		
		//	Create a list of empty arrays that have the right sizes.
		FastTable<double[]> arrs = FastTable.newInstance();
		for (int i=0; i < hd.nParams; ++i) {
			arrs.add(new double[hd.mrows]);
		}
		
		//	Read in the 2D array from the PROFIL file.
		readArray(hd, arrs, input);
		
        //  Convert from simple arrays to ArrayParam instances.
        FastTable<ArrayParam> arrParams = FastTable.newInstance();
		for (int i=0; i < hd.nParams; ++i) {
			ArrayParam array = ArrayParam.valueOf(paramNames.get(i), Dimensionless.UNIT, arrs.get(i));
			arrParams.add(array); 
		}
		FastTable.recycle(paramNames);
        
		//	PROFIL files contain only a single case, so create that.
		DataCase aCase = DataCase.newInstance(_defaultSetName);
		aCase.addAll(arrParams);
		FastTable.recycle(arrs);
        FastTable.recycle(arrParams);
		
		//	PROFIL files contain only a single data set, so create that.
		DataSet set = DataSet.newInstance(_defaultSetName);

		//	Add the case to the set.
		set.add(aCase);

		//	Add the set to the list of data sets being output.
		dataSets.add(set);

		//	Clean up before leaving.
		FastTable.recycle(hd.numBlocks);
		FastTable.recycle(hd.numSteps);
		
		return dataSets;
	}

	/**
	*  Method that writes out the data stored in the specified
	*  {@link DataSet} object to the specified output stream in POST II PROFIL format.
	*  This method will throw an exception as writing to this format is not yet supported.
	*
	*  @param  output  The output stream to which the data is to be written.
	*  @param  data    A list of {@link DataSet} objects containing data to be written out.
	*  @throws IOException If there is a problem writing to the specified stream.
	*/
    @Override
	public void write( OutputStream output, List<DataSet> data ) throws IOException {
		throw new IOException("This method has not yet been implemented");
	}


	/**
	*  Read in the 13312 byte header data and check for valid numbers.
	*
	*  @param is  PROFIL binary formatted input stream.
	*  @return The header data extracted from the input stream.
	*/
	private HeaderData readHeaderData(InputStream is) throws IOException {
	
		//	Create a header data object.
		HeaderData hd = new HeaderData();
		
		//	Allocate memory for storing a single block of the file.
		byte[] bytes = ArrayFactory.BYTES_FACTORY.array(BLOCK_SIZE);
		
		try {
			//	Read in the 1st block which contains the header info.
			int numRead = is.read(bytes, 0, BLOCK_SIZE);
			if (numRead != BLOCK_SIZE)	throw new IOException("Could not read the PROFIL file header.");
		
		
			//	Determine the byte order by looking at the 2 byte integer at the start of the file.
			ByteBuffer buffer = ByteBuffer.wrap(bytes);
			buffer = buffer.order(ByteOrder.LITTLE_ENDIAN);	//	Start by assuming it is Little Endian.
			int num = buffer.getShort();
			if (num == 25651)
				hd.byteOrder = ByteOrder.LITTLE_ENDIAN;
			else {
				hd.byteOrder = ByteOrder.BIG_ENDIAN;
				buffer.order(ByteOrder.BIG_ENDIAN);
			}
		
			//	Read in the identification record by converting the bytes to a string.
			BufferedReader reader = new BufferedReader(new InputStreamReader(new ByteArrayInputStream(bytes, 2, 7)));
			String name = reader.readLine();
			if (!name.equals("post II"))
				throw new IOException("Stream not recognized as containing POST II data.");
			
			//	Read in the number of phases.
			buffer.position(288);
			int nPhases = (int)buffer.getDouble();
			if (nPhases < 1)	throw new IOException("Incorrect number of phases.  Must be > 0, got: " + nPhases);
			hd.nPhases = nPhases;
			
			//	Read in the number of rows.
			int mrows = (int)buffer.getDouble();
			if (mrows < 1)	throw new IOException("Incorrect number of time steps in the data.  Must be > 0, got: " + mrows);
			hd.mrows = mrows;
			
			//	Skip in an unknown number.
			buffer.getDouble();
			
			//	Loop over all the phases.
			for (int i=0; i < nPhases; ++i) {
				//	Read in the phase ID.
				int phaseID = (int)buffer.getDouble();
				
				//	Skip the start time for the phase.
				buffer.getDouble();
				
				//	Skip an unknown number.
				buffer.getDouble();
				
				//	Read in the number of parameters in the phase.
				int nParams = (int)buffer.getDouble();
				if (nParams < 1)
					throw new IOException("Must have at least one parameter per phase, got: " + nParams +
											", for phase " + phaseID + ".");
				if (i == 0)
					hd.nParams = nParams;
				else if (nParams != hd.nParams)
					throw new IOException("All phases must have the same set of parameters. Phase " + phaseID +
											" has " + nParams + " parameters.");
				
				//	Read in the number of blocks used to hold the data for this phase.
				int nBlocks = (int)buffer.getDouble();
				if (nBlocks < 1)
					throw new IOException("Number of blocks must be > 0 for phase " + phaseID + ", got: " + nBlocks + ".");
				hd.numBlocks.add(Integer64.valueOf(nBlocks));
				
				//	Read in the number of time steps (rows) in this phase.
				int numSteps = (int)buffer.getDouble();
				if (numSteps < 1)
					throw new IOException("Number of steps must be > 0 for phase " + phaseID + ", got: " + numSteps + ".");
				hd.numSteps.add(Integer64.valueOf(numSteps));
			}
			
			
		} catch (BufferUnderflowException e) {
			//	Shouldn't be possible.
			FastTable.recycle(hd.numBlocks);
			FastTable.recycle(hd.numSteps);
			e.printStackTrace();
			throw new IOException(e.getMessage());
			
		} catch (IndexOutOfBoundsException e) {
			FastTable.recycle(hd.numBlocks);
			FastTable.recycle(hd.numSteps);
			e.printStackTrace();
			throw new IOException(e.getMessage());
			
		} finally {
			ArrayFactory.BYTES_FACTORY.recycle(bytes);
		}
		
		if (DEBUG) {
			System.out.println("hd.byteOrder = " + hd.byteOrder);
			System.out.println("hd.nPhases = " + hd.nPhases);
			System.out.println("hd.mrows = " + hd.mrows);
			System.out.println("hd.nParams = " + hd.nParams);
			
			for (int i=0; i < hd.nPhases; ++i) {
				System.out.println("hd.numBlocks(" + i + ") = " + hd.numBlocks.get(i) +
					", hd.numSteps(" + i + ") = " + hd.numSteps.get(i));
			}
		}
		
		return hd;
	}
	
	
	/**
	*  Method that reads in the names of the parameters from the POST II PROFIL file and
	*  returns a list of Text objects.
	*
	*  @param hd  The header data extracted from the input stream.
	*  @param is  POST PROFIL binary formatted input stream.
	*  @return A String containing the name of the array in the file.
	*/
	private FastTable<Text> readParamNames(HeaderData hd, InputStream is) throws IOException {
		FastTable<Text> names = FastTable.newInstance();
		
		//	Allocate memory for a single block of data from the file.
		byte[] bytes = ArrayFactory.BYTES_FACTORY.array(BLOCK_SIZE);
		
		try {
			//	Read in the 2nd block which contains the parameter names.
			int numRead = is.read(bytes, 0, BLOCK_SIZE);
			if (numRead != BLOCK_SIZE)	throw new IOException("Could not read the PROFIL parameter names block.");
			
			//	Loop over all the parameters.
			int pos = 296;
			for (int i=0; i < hd.nParams; ++i) {
			
				//	Convert the bytes to a String.
				BufferedReader reader = new BufferedReader(new InputStreamReader(new ByteArrayInputStream(bytes, pos, 24)));
				String name = reader.readLine();
				if (name == null)	throw new IOException("Trouble reading the parameter name from the PROFIL file.");
				names.add(Text.valueOf(name).trim());
				
				if (DEBUG) {
					System.out.println(" name = " + name.trim());
				}
				pos = pos + 24;
			}
			
		} finally {
			ArrayFactory.BYTES_FACTORY.recycle(bytes);
		}
		
		return names;
	}
	
	/**
	*  Method that reads in the array of numbers from the PROFIL file and
	*  returns a 2D array of double values.
	*
	*  @param hd   The header data extracted from the input stream.
	*  @param arrs A list of arrays already sized to hold the data (data will be filled in by this method).
	*  @param is   PROFIL binary formatted input stream with the 1st 2 blocks already read in.
	*/
	private void readArray(HeaderData hd, List<double[]> arrs, InputStream is) throws IOException {
			
		//	Allocate memory for storing a single block of the file.
		byte[] bytes = ArrayFactory.BYTES_FACTORY.array(BLOCK_SIZE);
		
		try {
			int currentRow = 0;
			
			//	Loop over all the phases.
			for (int phaseIdx = 0; phaseIdx < hd.nPhases; ++phaseIdx) {
				if (phaseIdx != 0) {
					//	Skip the parameter name block.
					is.read(bytes, 0, BLOCK_SIZE);
				}
				
				//	Get the number of blocks and steps required.
				int numBlocks = hd.numBlocks.get(phaseIdx).intValue();
				int numRows = hd.numSteps.get(phaseIdx).intValue();
				
				if (DEBUG) {
					System.out.println("phase = " + phaseIdx);
					System.out.println("numBlocks = " + numBlocks);
					System.out.println("numRows = " + numRows);
				}
				
				//	Loop over all the blocks to be read in for this phase.
				for (int blockIdx = 0; blockIdx < numBlocks; ++blockIdx) {
				
					//	Read in the next block of data.
					int numRead = is.read(bytes, 0, BLOCK_SIZE);
					if (numRead != BLOCK_SIZE)	throw new IOException("Unexpected end of file encountered in PROFIL.");
			
					//	Create a byte buffer to translate the bytes.
					ByteBuffer buffer = ByteBuffer.wrap(bytes);
					buffer = buffer.order(hd.byteOrder);			//	The buffer will convert byte order if necessary.
				
					//	Translate the bytes into arrays of double values.
					int mrows = translateBytes(hd, currentRow, numRows, arrs, buffer.asDoubleBuffer());
					currentRow += mrows;
					numRows -= mrows;
				}
				
			}
			
		} finally {
			ArrayFactory.BYTES_FACTORY.recycle(bytes);
		}
		
	}
	
	
	/**
	*  Translate a ByteBuffer of doubles into an array of doubles.  Returns the number of rows (steps)
	*  read in from the file.
	*
	*  @param hd        The header data extracted from the input stream.
	*  @param rowOffset The offset into the output arrays of the next row to be read in.
	*  @param mrows		The desired number of rows to be read in for this case (only the number that actually
	*					fit in this block will actually be read in).
	*  @param arrs A list of arrays already sized to hold the data (data will be filled in by this method).
	*  @param buffer	A buffer containing doubles for the current data block.
	*  @return The number of rows actually read in by this method.  Will always be <= the number of rows that
	*			fit in this data block.
	*/
	private int translateBytes(HeaderData hd, int rowOffset, int mrows, List<double[]> arrs, DoubleBuffer buffer) {
		
		//	Extract the number of rows and columns.
		int ncols = hd.nParams;
		
		//	Only read what will fit in one block.
		mrows = Math.min(mrows, BLOCK_SIZE/(ncols*8));
		
		//	Translate the data from the buffer.
		for (int i=0; i < mrows; ++i) {
			for (int j=0; j < ncols; ++j) {
				arrs.get(j)[rowOffset] = buffer.get();
			}
			++rowOffset;
		}
		
		return mrows;
	}
	
	
	/**
	*  Class that contains the header data for the MatLab V4 file.
	*/
	private class HeaderData {
		//	The byte order of the input stream.
		ByteOrder byteOrder;
		
		//	Number of phases in the data.
		int nPhases;
		
		//	Number of rows in the data (number of time steps).
		int mrows;
		
		//	Number of parameters.
		int nParams;
		
		//	Number of blocks in each phase.
		FastTable<Integer64> numBlocks = FastTable.newInstance();
		
		//	Number of time steps in each phase.
		FastTable<Integer64> numSteps = FastTable.newInstance();
	}

}



