/**
 * SelectCasesDialog -- Dialog allowing user to select a subset of a list of cases.
 *
 * Copyright (C) 2003-2016, by Joseph A. Huwaldt. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA. Or visit: http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.datareader;

import jahuwaldt.swing.EscapeJDialog;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.*;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;
import javax.swing.*;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.table.TableRowSorter;

/**
 * A modal dialog the allows the user to select a subset of all the available cases in a
 * list of data sets.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt, Date: March 25, 2003
 * @version October 29, 2016
 */
@SuppressWarnings("serial")
public class SelectCasesDialog extends EscapeJDialog implements ActionListener {

    /**
     * The resource bundle for this package.
     */
    public static final ResourceBundle RESOURCES = DataReader.RESOURCES;
    
    //  A list of column names to display above the table.
    private static final String[] columnNames = {
        padRight(RESOURCES.getString("dataSetColName"), 20),
        padRight(RESOURCES.getString("dataCaseColName"), 20)};

    //  A list of lists of strings.
    private List<List<Object>> rowData = new ArrayList<>();

    /**
     * The full list of DataElements displayed to the user.
     */
    private List<DataElement> elements = new ArrayList<>();

    /**
     * The total number of input DataSets.
     */
    private int numSets = 0;

    /**
     * The JTable displaying the contents of the "elements" list that was input.
     */
    private JTable inputTable;

    /**
     * The row sorter used on the input table.
     */
    private TableRowSorter<AbstractTableModel> inputSorter;

    /**
     * The filter text field for filtering the input table.
     */
    private final JTextField filterText;

    /**
     * The JTable displaying the elements that the user has selected. This table is only
     * used if multiple selections are possible.
     */
    private JTable outputTable;

    //  A list of the cases that were selected by the user.
    private DataSet selectedCases;

    /**
     * Flag indicating if the user is allowed to select multiple items or not.
     */
    private boolean singleSelection = false;

    //  A list of lists of selected elements.
    private ArrayList<ArrayList<Object>> selRowData = new ArrayList<>();
    private ArrayList<DataElement> selElements = new ArrayList<>();

    /**
     * Construct a model dialog that allows the user to select a subset of all the
     * available cases in a list of data sets.
     *
     * @param parent          A reference to the parent frame that this dialog belongs to.
     * @param title           The title for the dialog window.
     * @param message         A message to display telling the user what to do.
     * @param data            A list of DataSet objects containing the runs to be chosen
     *                        from.
     * @param singleSelection Set to true to allow only a single element to be selected,
     *                        false for multiple elements.
     * @param showArraySizes  Append the size of ArrayParam objects to the parameter names
     *                        if set to true. Show only the parameter names if set to
     *                        false.
     */
    public SelectCasesDialog(Frame parent, String title, String message, List<? extends DataElementList<?>> data,
            boolean singleSelection, boolean showArraySizes) {
        super(parent, title, true);

        //  Copy all the cases from the input data sets into the list.
        numSets = data.size();
        for (DataElementList<?> list : data) {

            for (DataElement element : list) {
                elements.add(element);
                ArrayList<Object> colData = new ArrayList<>();
                colData.add(list.getName().toString());
                if (element instanceof ArrayParam) {
                    String str = element.toString();
                    if (showArraySizes)
                        str += "  (" + ((ArrayParam)element).size() + ")";
                    colData.add(str);
                } else
                    colData.add(element);
                rowData.add(colData);
            }
        }

        //  If there is only one row, simplify the dialog.
        if (rowData.size() == 1)
            singleSelection = true;
        this.singleSelection = singleSelection;

        //  Layout the dialog window.
        JPanel cp = new JPanel(new BorderLayout());
        this.getContentPane().add(cp);
        cp.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));

        //  Add instructions at the top.
        Box topPanel = Box.createHorizontalBox();
        cp.add(topPanel, BorderLayout.NORTH);

        JLabel label = new JLabel(message, SwingConstants.LEADING);
        label.setBorder(BorderFactory.createEmptyBorder(6, 0, 6, 0));
        topPanel.add(label);
        topPanel.add(Box.createHorizontalStrut(10));
        topPanel.add(Box.createGlue());

        //  Create a panel to display in the center.
        Box centerPanel = Box.createVerticalBox();
        cp.add(centerPanel, BorderLayout.CENTER);
        Box tablePanel = Box.createHorizontalBox();
        centerPanel.add(tablePanel);

        //  Create a table that displays the cases in all the input data sets.
        AbstractTableModel model = new AbstractTableModel() {
            @Override
            public String getColumnName(int col) {
                if (numSets == 1)
                    col = 1;
                return SelectCasesDialog.this.getColumnName(col);
            }

            @Override
            public int getRowCount() {
                return rowData.size();
            }

            @Override
            public int getColumnCount() {
                if (numSets == 1)
                    return 1;
                return columnNames.length;
            }

            @Override
            public Object getValueAt(int row, int col) {
                if (numSets == 1)
                    col = 1;
                return ((ArrayList<?>)rowData.get(row)).get(col);
            }

            @Override
            public boolean isCellEditable(int row, int col) {
                return false;
            }
        };
        inputTable = new JTable(model);
        inputSorter = new TableRowSorter<>(model);
        inputTable.setRowSorter(inputSorter);
        inputTable.setCellSelectionEnabled(false);
        inputTable.setColumnSelectionAllowed(false);
        inputTable.setRowSelectionAllowed(true);
        inputTable.clearSelection();
        inputTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);   //  Show horiz. scroll bars.
        inputTable.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                if (e.getClickCount() == 2) {
                    if (!SelectCasesDialog.this.singleSelection)
                        doAdd();
                    else
                        SelectCasesDialog.this.actionPerformed(null);
                }
            }
        });
        if (singleSelection)
            inputTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        //  If only a single row exists, preselect it.
        if (rowData.size() == 1)
            inputTable.selectAll();
        else {
            ListSelectionModel selModel = inputTable.getSelectionModel();
            selModel.setSelectionInterval(0, 0);
        }

        //  Initialize the column widths to show all the data.
        initColumnWidths(inputTable);

        //  Determine the table preferred width.
        int width = 0;
        int numColumns = inputTable.getColumnCount();
        for (int i = 0; i < numColumns; ++i) {
            TableColumn column = inputTable.getColumnModel().getColumn(i);
            width += column.getPreferredWidth();
        }
        inputTable.setPreferredScrollableViewportSize(new Dimension(width, 300));

        //  Put the list in a scroll pane so that we have scroll bars.
        JScrollPane scrollPane = new JScrollPane(inputTable);
        scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);

        //  Add the list to the panel.
        tablePanel.add(scrollPane);

        if (!singleSelection) {

            //  Create a button panel for selection buttons.
            tablePanel.add(Box.createHorizontalStrut(10));
            tablePanel.add(createButtonPanel());

            //  Create an output table that displays selected tables.
            outputTable = new JTable(new AbstractTableModel() {
                @Override
                public String getColumnName(int col) {
                    if (numSets == 1)
                        col = 1;
                    return SelectCasesDialog.this.getColumnName(col);
                }

                @Override
                public int getRowCount() {
                    return selRowData.size();
                }

                @Override
                public int getColumnCount() {
                    if (numSets == 1)
                        return 1;
                    return columnNames.length;
                }

                @Override
                public Object getValueAt(int row, int col) {
                    if (numSets == 1)
                        col = 1;
                    return ((ArrayList<?>)selRowData.get(row)).get(col);
                }

                @Override
                public boolean isCellEditable(int row, int col) {
                    return false;
                }
            });
            outputTable.setCellSelectionEnabled(false);
            outputTable.setColumnSelectionAllowed(false);
            outputTable.setRowSelectionAllowed(true);
            outputTable.clearSelection();
            outputTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);  //  Show hoirz. scroll bars.
            outputTable.addMouseListener(new MouseAdapter() {
                @Override
                public void mouseClicked(MouseEvent e) {
                    if (e.getClickCount() == 2) {
                        doRemove();
                    }
                }
            });

            //  Initialize the column widths to show all the data.
            initColumnWidths(outputTable);

            //  Determine the table preferred width.
            width = 0;
            numColumns = outputTable.getColumnCount();
            for (int i = 0; i < numColumns; ++i) {
                TableColumn column = outputTable.getColumnModel().getColumn(i);
                width += column.getPreferredWidth();
            }
            outputTable.setPreferredScrollableViewportSize(new Dimension(width, 300));

            //  Put the list in a scroll pane so that we have scroll bars.
            scrollPane = new JScrollPane(outputTable);
            scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
            scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);

            //  Add the list to the panel.
            tablePanel.add(Box.createHorizontalStrut(10));
            tablePanel.add(scrollPane);
        }

        //  Define the filter text field.
        centerPanel.add(Box.createVerticalStrut(10));
        Box form = Box.createHorizontalBox();
        JLabel l1 = new JLabel(RESOURCES.getString("filterTextLabel"), SwingConstants.TRAILING);
        form.add(l1);
        filterText = new JTextField();
        //  Whenever filterText changes, invoke newInputFilter.
        filterText.getDocument().addDocumentListener(new DocumentListener() {
            @Override
            public void changedUpdate(DocumentEvent e) {
                newInputFilter();
            }

            @Override
            public void insertUpdate(DocumentEvent e) {
                newInputFilter();
            }

            @Override
            public void removeUpdate(DocumentEvent e) {
                newInputFilter();
            }
        });
        l1.setLabelFor(filterText);
        form.add(filterText);
        centerPanel.add(form);
        centerPanel.add(Box.createVerticalStrut(10));

        //  Define OK and Cancel buttons.
        Box box = Box.createHorizontalBox();
        box.add(Box.createGlue());

        JButton cancelBtn = new JButton(RESOURCES.getString("cancelBtnText"));
        cancelBtn.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                performEscapeAction(null);
            }
        });
        box.add(cancelBtn);

        box.add(Box.createHorizontalStrut(40));

        //  If multiple selections allowed, change "OK" button text to "Done".
        String okMessage = RESOURCES.getString("doneBtnText");
        if (singleSelection)
            okMessage = RESOURCES.getString("selectBtnText");
        JButton okayBtn = new JButton(okMessage);
        okayBtn.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                SelectCasesDialog.this.actionPerformed(null);
            }
        });
        this.getRootPane().setDefaultButton(okayBtn);
        box.add(okayBtn);

        box.add(Box.createHorizontalStrut(20));
        cp.add(box, BorderLayout.SOUTH);

        this.pack();
    }

    /**
     * Pad a string by adding spaces on the right side until the given length is reached.
     *
     * @param s The string to be padded.
     * @param n The desired length of the string.
     * @return The input string with spaces added to the right until the given length is
     *         reached
     */
    protected static String padRight(String s, int n) {
        return String.format("%1$-" + n + "s", s);
    }
    
    /**
     * This method returns a single DataSet object that contains all the cases that were
     * selected in this dialog. If the user has not clicked on the "OK" button, null is
     * returned.
     *
     * @return A single DataSet object that contains all the cases that were selected in
     *         this dialog.
     */
    public DataSet getSelected() {
        return selectedCases;
    }

    /**
     * Method that creates a button panel containing selection buttons.
     */
    private Box createButtonPanel() {

        //  Create selection buttons to run down the center.
        Box btnBox = Box.createVerticalBox();
        btnBox.add(Box.createGlue());

        //  Create an "Add" button.
        JButton button = new JButton(RESOURCES.getString("addBtnText"));
        button.setAlignmentX(JButton.CENTER_ALIGNMENT);
        button.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doAdd();
            }
        });
        btnBox.add(button);
        btnBox.add(Box.createVerticalStrut(10));

        //  Create a "Remove" button.
        button = new JButton(RESOURCES.getString("removeBtnText"));
        button.setAlignmentX(JButton.CENTER_ALIGNMENT);
        button.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doRemove();
            }
        });
        btnBox.add(button);
        btnBox.add(Box.createVerticalStrut(10));

        //  Create an "Add All" button.
        button = new JButton(RESOURCES.getString("addAllBtnText"));
        button.setAlignmentX(JButton.CENTER_ALIGNMENT);
        button.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doAddAll();
            }
        });
        btnBox.add(button);
        btnBox.add(Box.createVerticalStrut(10));

        //  Create a "Remove All" button.
        button = new JButton(RESOURCES.getString("removeAllBtnText"));
        button.setAlignmentX(JButton.CENTER_ALIGNMENT);
        button.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doRemoveAll();
            }
        });
        btnBox.add(button);
        btnBox.add(Box.createGlue());

        return btnBox;
    }

    /**
     * This method is called when the user clicks on the "Add" button.
     */
    private void doAdd() {

        //  Copy the selected elements to the output table.
        int[] selected = inputTable.getSelectedRows();
        int numSelected = selected.length;

        for (int i = 0; i < numSelected; ++i) {
            int modelRow = inputTable.convertRowIndexToModel(selected[i]);  //  Convert from View to Model indexes.
            DataElement element = elements.get(modelRow);
            selElements.add(element);
            ArrayList<Object> colData = new ArrayList<>();
            colData.add(inputTable.getValueAt(selected[i], 0));
            colData.add(inputTable.getValueAt(selected[i], 1));
            selRowData.add(colData);
        }

        if (numSelected > 0)
            //  Update the output table's UI.
            outputTable.revalidate();

    }

    /**
     * This method is called when the user clicks on the "Add All" button.
     */
    private void doAddAll() {

        int numRows = inputTable.getRowCount();
        for (int i = 0; i < numRows; ++i) {
            int modelRow = inputTable.convertRowIndexToModel(i);    //  Convert from View to Model indexes.
            DataElement element = elements.get(modelRow);
            selElements.add(element);
            ArrayList<Object> colData = new ArrayList<>();
            colData.add(inputTable.getValueAt(i, 0));
            colData.add(inputTable.getValueAt(i, 1));
            selRowData.add(colData);
        }

        outputTable.revalidate();

    }

    /**
     * This method is called when the user clicks on the "Remove" button.
     */
    private void doRemove() {

        //  Get a list of all the selected rows in the table.
        int[] selected = outputTable.getSelectedRows();
        int numSelected = selected.length;

        for (int i = 0; i < numSelected; ++i) {
            selElements.remove(selected[i]);
            selRowData.remove(selected[i]);
        }

        outputTable.revalidate();
    }

    /**
     * This method is called when the user clicks on the "Remove All" button.
     */
    private void doRemoveAll() {
        selElements.clear();
        selRowData.clear();

        outputTable.revalidate();
    }

    /**
     * Method that returns a list of selected elements.
     * 
     * @return A list of selected elements.
     */
    protected DataElement[] getSelectedElements() {

        DataElement[] selList;

        if (singleSelection) {
            //  Get a list of all the selected rows in the input table.
            int[] selected = inputTable.getSelectedRows();
            int numSelected = selected.length;

            selList = new DataElement[numSelected];
            for (int i = 0; i < numSelected; ++i) {
                int modelRow = inputTable.convertRowIndexToModel(selected[i]);  //  Convert from View to Model indexes.
                selList[i] = elements.get(modelRow);
            }
        } else {
            //  Return a list of all the elements in the output table.
            int numRows = outputTable.getRowCount();
            selList = new DataElement[numRows];
            for (int i = 0; i < numRows; ++i) {
                selList[i] = selElements.get(i);
            }

        }

        return selList;
    }

    /**
     * This method is called when the user clicks on the "OK Button".
     * 
     * @param evt The event that caused this method to be called (ignored).
     */
    @Override
    public void actionPerformed(ActionEvent evt) {

        //  Get a list of all the selected rows in the table.
        DataElement[] selected = getSelectedElements();
        int numSelected = selected.length;

        //  Add selected cases to the selected data set.
        selectedCases = DataSet.newInstance(RESOURCES.getString("selectedSetName"));
        for (int i = 0; i < numSelected; ++i) {
            selectedCases.add((DataCase)selected[i]);
        }

        //  Hide the window.
        this.setVisible(false);
    }

    /**
     * Response to ENTER key pressed goes here. This has the same result as clicking on
     * the OK or Select button.
     */
    @Override
    protected void performEnterAction(KeyEvent e) {
        actionPerformed(null);
    }

    /**
     * Method that returns the name to place above the specified column.
     *
     * @param col The column to return the name for.
     * @return The name to use for the specified column.
     */
    protected String getColumnName(int col) {
        return columnNames[col];
    }

    /**
     * Method that resets the minimum column widths.
     */
    private static void initColumnWidths(JTable table) {
        int numColumns = table.getColumnCount();
        for (int i = 0; i < numColumns; ++i) {
            //  Get each column.
            TableColumn column = table.getColumnModel().getColumn(i);

            //  Determine the header width.
            int headerWidth = 0;
            try {
                TableCellRenderer renderer = column.getHeaderRenderer();
                if (renderer == null)
                    renderer = table.getTableHeader().getDefaultRenderer();

                Component comp = renderer.
                        getTableCellRendererComponent(
                                null, column.getHeaderValue(),
                                false, false, 0, 0);
                headerWidth = comp.getPreferredSize().width;
            } catch (NullPointerException e) {
                e.printStackTrace();
            }

            //  Set the minimum width to the larger of the header width
            //  and 150 pixels.
            column.setMinWidth(Math.max(150, headerWidth) + 10);
        }
    }

    /**
     * Method called whenever the filter text is edited to change the input table filter.
     */
    private void newInputFilter() {
        RowFilter<AbstractTableModel, Object> rf;
        //  If current expression doesn't parse, don't update.
        try {
            rf = RowFilter.regexFilter(filterText.getText(), 0);
        } catch (java.util.regex.PatternSyntaxException e) {
            return;
        }
        inputSorter.setRowFilter(rf);

        //  Make sure something in the table is selected.
        if (inputTable.getSelectedRow() < 0) {
            ListSelectionModel selModel = inputTable.getSelectionModel();
            selModel.setSelectionInterval(0, 0);
        }
    }
}
