/**
 * SelectParamsDialog -- Dialog allowing user to select a subset from a list of
 * parameters.
 *
 * Copyright (C) 2003-2015, by Joseph A. Huwaldt. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA. Or visit: http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.datareader;

import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.util.List;

/**
 * A modal dialog the allows the user to select a subset of all the available parameters
 * in a list of data cases.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt, Date: May 7, 2003
 * @version October 15, 2015
 */
public class SelectParamsDialog extends SelectCasesDialog {

    //  A list of column names to display above the table.
    private static final String[] columnNames = {
        padRight(RESOURCES.getString("dataCaseColName"), 20),
        padRight(RESOURCES.getString("paramColName"), 20)
    };
    
    //  A list of the parameters that were selected by the user.
    private DataSet selectedParams;
    
    //  The list of data cases we are selecting from.
    private final List<DataCase> allCases;

    /**
     * Construct a model dialog that allows the user to select a subset of all the
     * available cases in a list of data sets.
     *
     * @param parent          A reference to the parent frame that this dialog belongs to.
     * @param title           The title for the dialog window.
     * @param message         A message to display telling the user what to do.
     * @param data            A list of DataCase objects containing the parameters to be
     *                        chosen from.
     * @param singleSelection Set to true to allow only a single parameter to be selected,
     *                        false for multiple parameters.
     * @param showArraySizes  Append the size of ArrayParam objects to the parameter names
     *                        if set to true. Show only the parameter names if set to
     *                        false.
     */
    public SelectParamsDialog(Frame parent, String title, String message, List<DataCase> data,
            boolean singleSelection, boolean showArraySizes) {
        super(parent, title, message, data, singleSelection, showArraySizes);

        allCases = data;

    }

    /**
     * This method returns a single DataSet object that contains all the parameters that
     * were selected in this dialog stored in a set of new DataCase objects (one parameter
     * per case). If the user has not clicked on the "OK" button, null is returned.
     */
    @Override
    public DataSet getSelected() {
        return selectedParams;
    }

    /**
     * This method is called when the user clicks on the "OK Button".
     */
    @Override
    public void actionPerformed(ActionEvent evt) {

        //  Create a new data set to hold the cases we are creating to
        //  hold the selected parameters.
        selectedParams = DataSet.newInstance(RESOURCES.getString("selectedSetName"));

        //  Get a list of all the selected rows in the table.
        DataElement[] selected = getSelectedElements();
        int numSelected = selected.length;

        //  Add selected parameters to a new DataCase that corresponds
        //  to the one the parameter originally came from.
        int numCases = allCases.size();
        for (int i = 0; i < numSelected; ++i) {
            DataParam param = (DataParam)selected[i];

            //  Figure out what case this parameter is from.
            int caseID = 0;
            DataCase parentCase = null;
            for (; caseID < numCases; ++caseID) {
                parentCase = (DataCase)allCases.get(caseID);
                if (parentCase.contains(param))
                    break;
            }

            //  Do we already have a case with this name?
            boolean found = false;
            DataCase selCase = null;
            int selCases = selectedParams.size();
            for (int j = 0; j < selCases; ++j) {
                selCase = (DataCase)selectedParams.get(j);
                CharSequence parentCaseName = (parentCase == null ? null : parentCase.getName());
                if (selCase.getName().equals(parentCaseName)) {
                    found = true;
                    break;
                }
            }

            //  If we already have this case, add the parameter to it.
            if (found && selCase != null)
                selCase.add(param);
            else {
                //  Create a new case having the same name as the case where this
                //  parameter originally came from.
                CharSequence parentCaseName = (parentCase == null ? null : parentCase.getName());
                DataCase tempCase = DataCase.newInstance(parentCaseName);
                tempCase.add(param);

                selectedParams.add(tempCase);
            }
        }

        //  Hide the window.
        this.setVisible(false);
    }

    /**
     * Method that returns the name to place above the specified column.
     */
    @Override
    protected String getColumnName(int col) {
        return columnNames[col];
    }
}
