/**
 * TREETOPSDataReader -- Reads data from NASA TREETOPS/CLVTOPS simulation data files.
 *
 * Copyright (C) 2017-2024, by Joseph A. Huwaldt. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA. Or visit: http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.datareader;

import jahuwaldt.io.FileUtils;
import jahuwaldt.js.unit.EditUnitSetDialog;
import jahuwaldt.js.unit.UnitSet;
import jahuwaldt.swing.AppUtilities;
import java.awt.Frame;
import java.io.*;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.MessageFormat;
import java.util.*;
import javax.measure.quantity.AngularVelocity;
import javax.measure.quantity.Dimensionless;
import javax.measure.unit.SI;
import javax.measure.unit.Unit;

/**
 * This class reads data from a NASA TREETOPS/CLVTOPS *.dat and matching *.crf file and
 * returns a list of DataSet data structures containing the simulation data.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt, Date: March 17, 2017
 * @version March 27, 2024
 */
public class TREETOPSDataReader implements DataReader {

    //  A brief description of the data read by this reader.
    private static final String DESCRIPTION = RESOURCES.getString("treetopsDesc");

    //  The preferred file extension for files of this reader's type.
    public static final String EXTENSION = RESOURCES.getString("treetopsDefExt");

    //  Define constants for each element type.
    private static final String BODY = "Body";
    private static final String HINGE = "Hinge";
    private static final String SENSOR = "Sensor";
    private static final String FUNCTION = "Function";
    private static final String ACTUATOR = "Actuator";
    private static final String DEVICE = "Device";
    private static final String CONTROLLER_OUT = "Controller Output";
    private static final String CONTROLLER_IN = "Controller Input";
    
    //  Define dictionaries containing the output variable names, descriptions and units.
    private static final HashMap<String,String[]> OUT_VARS = new HashMap();
    private static final HashMap<String,String[]> OUT_DESC = new HashMap();
    private static final HashMap<String,String[]> OUT_UNITS = new HashMap();
    
    //  Define a dictionary mapping the element type to it's abbreviation.
    private static final HashMap<String,String> IDABR = new HashMap();

    static {
        //  Define the body output variable names, descriptions and units.
        String[] bodyOutputVars = RESOURCES.getString("ttBodyOutputVars").split(",");
        String[] bodyOutputDesc = RESOURCES.getString("ttBodyOutputDesc").split(",");
        String[] bodyOutputUnits = RESOURCES.getString("ttBodyOutputUnits").split(",");
        
        //  Store in the appropriate dictionaries.
        OUT_VARS.put(BODY, bodyOutputVars);
        OUT_DESC.put(BODY, bodyOutputDesc);
        OUT_UNITS.put(BODY, bodyOutputUnits);

        //  Define the hinge output variable names, descriptions and units.
        String[] hingeOutputVars = RESOURCES.getString("ttHingeOutputVars").split(",");
        String[] hingeOutputDesc = RESOURCES.getString("ttHingeOutputDesc").split(",");
        String[] hingeOutputUnits = RESOURCES.getString("ttHingeOutputUnits").split(",");

        //  Store in the appropriate dictionaries.
        OUT_VARS.put(HINGE, hingeOutputVars);
        OUT_DESC.put(HINGE, hingeOutputDesc);
        OUT_UNITS.put(HINGE, hingeOutputUnits);

        //  Define the element type ID abbreviations.
        IDABR.put(BODY, "BO");
        IDABR.put(HINGE, "HI");
        IDABR.put(SENSOR, "SE");
        IDABR.put(FUNCTION, "FU");
        IDABR.put(ACTUATOR, "AC");
        IDABR.put(DEVICE, "DEMAG");
        IDABR.put(CONTROLLER_OUT, "RC");
        IDABR.put(CONTROLLER_IN, "UC");
    }

    //  Since the TREETOPS data is stored in a space delimited array, we can just use
    //  The tab data reader to process the *.dat file.
    private final DataReader datReader = new TabDataReader();

    /**
     * Returns a string representation of the object. This will return a brief description
     * of the format read by this reader.
     */
    @Override
    public String toString() {
        return DESCRIPTION;
    }

    /**
     * Returns the preferred file extension (not including the ".") for files of this
     * DataReader's type.
     *
     * @return The preferred file extension for this file's type.
     */
    @Override
    public String getExtension() {
        return EXTENSION;
    }

    /**
     * Compares this object with the specified object for order based on the
     * <code>toString().compareTo(o.toString())</code> method. Returns a negative integer,
     * zero, or a positive integer as this object is less than, equal to, or greater than
     * the specified object.
     */
    @Override
    public int compareTo(DataReader o) {
        return this.toString().compareTo(o.toString());
    }

    /**
     * Method that determines if this reader can read data from the specified input
     * stream.
     *
     * @param pathName The path to the file to be read.
     * @param input    An input stream containing the data to be read. Any methods that
     *                 read from this stream must first set a mark and then reset back to
     *                 that mark before the method returns (even if it returns with an
     *                 exception).
     * @return DataReader.NO if the file is not recognized at all or DataReader.MAYBE if
     * the file has an appropriate extension.
     * @throws java.io.IOException If the input stream could not be read from.
     */
    @Override
    public int canReadData(String pathName, BufferedInputStream input) throws IOException {
        int answer = DataReader.NO;

        pathName = pathName.trim();
        String prefix = FileUtils.getFileNameWithoutExtension(pathName);
        pathName = pathName.toLowerCase();
        
        File crfFile = new File(prefix + ".crf");
        if (crfFile.exists()) {
            if (pathName.endsWith(".dat")) {
                answer = DataReader.YES;
            }
        }
        
        return answer;
    }

    /**
     * Returns false. This class can not write data to a CLVTOPS formatted file stream.
     *
     * @return Always returns false.
     */
    @Override
    public boolean canWriteData() {
        return false;
    }

    /**
     * Method that reads in CLVTOPS formatted simulation data from the specified
     * input stream and returns that data as a list of {@link DataSet} objects.
     *
     * @param pathName The path to the file to be read.  In addition to the *.dat file that
     * this should point to, there must be a *.crf file with the same prefix in the same
     * directory that will also be read in.
     * @param input    An input stream containing the space delimited array data file (*.dat).
     * @return A list of DataSet objects that contains the data read in from the
     * specified stream (will contain a single data set which will contain a single case
     * with ArrayParam objects for each column).
     * @throws IOException If there is a problem reading the specified stream.
     */
    @Override
    public List<DataSet> read(String pathName, InputStream input) throws IOException {
        
        //  Define the units for the data.
        UnitSet nativeUnits = askForUnitSet();
        if (nativeUnits == null)
            return null;
        
        List<DataSet> data = read(pathName, input, nativeUnits);
        
        return data;
    }

   /**
     * Method that reads in CLVTOPS formatted simulation data from the specified
     * input stream and returns that data as a list of {@link DataSet} objects.
     *
     * @param pathName The path to the file to be read.  In addition to the *.dat file that
     * this should point to, there must be a *.crf file with the same prefix in the same
     * directory that will also be read in.
     * @param input    An input stream containing the space delimited array data file (*.dat).
     * @param nativeUnits The units for the data in the *.dat file.
     * @return A list of DataSet objects that contains the data read in from the
     * specified stream (will contain a single data set which will contain a single case
     * with ArrayParam objects for each column).
     * @throws IOException If there is a problem reading the specified stream.
     */
    public List<DataSet> read(String pathName, InputStream input, UnitSet nativeUnits) throws IOException {
        Objects.requireNonNull(nativeUnits);
        
        //  Start by reading in the tabular data from the input stream pointing to the
        //  *.dat file.
        List<DataSet> data = datReader.read(pathName, input);
        
        //  Now parse the CRF file to define the names for each column in the table.
        File crfFile = new File(FileUtils.getFileNameWithoutExtension(pathName) + ".crf");
        read(crfFile, data.get(0).get(0), nativeUnits);
        
        return data;
    }

    /**
     * Method that asks the user to select the units for the data being read in from the file.
     * 
     * @return The unit set for the data in the file.
     */
    private static UnitSet askForUnitSet() {
        
        //  Define the default unit set for the locale.
        UnitSet nativeUnits = new UnitSet(UnitSet.UnitSystem.SI_MKS);
        if (Locale.getDefault().equals(Locale.US))
            nativeUnits = new UnitSet(UnitSet.UnitSystem.US_FSS);
        
        //  Ask the user to select the units for the data in the file.
        EditUnitSetDialog dialog = new EditUnitSetDialog(null, RESOURCES.getString("editUnitSetTitle"),
                RESOURCES.getString("editUnitSetMsg"), nativeUnits);
        dialog.setLocation(AppUtilities.dialogPosition(dialog));
        dialog.setVisible(true);
        nativeUnits = dialog.getUnits();
        dialog.dispose();
        
        return nativeUnits;
    }
    
    /**
     * Read in the CLVTOPS CRF file and parse its contents.  The CRF file defines what
     * each column in the output data file is.
     * 
     * @param crf A reference to the *.crf file for the data file (*.dat) being read in.
     * @param rawData   The raw data read in from the *.dat file.
     * @param nativeUnits The units for the data in the *.dat file.
     * @return A reference to the input rawData after the names of each array have been
     * changed to reflect the data in the CRF file.
     * @throws IOException 
     */
    private static DataCase read(File crf, DataCase rawData, UnitSet nativeUnits) throws IOException {
        //  Open the CRF file and get a line number reader to it.
        try (FileInputStream is = new FileInputStream(crf)) {
            LineNumberReader reader = new LineNumberReader(new InputStreamReader(is));
            
            //  Parse the meta-data from the CRF file.
            parseCRFStream(reader, rawData, nativeUnits);
        }
        
        return rawData;
    }

    /**
     * Method that actually parses the CRF file.
     *
     * @param in          A reader that reads in the characters from the CRF file.
     * @param data        A list of all the ArrayParams as read from the tabular data file
     *                    (one ArrayParam per column in the CLVTOPS array). These
     *                    parameters will be modified by this method by adding names,
     *                    units, and reference data.
     * @param nativeUnits The system of units used by the data.
     */
    private static void parseCRFStream(LineNumberReader in, DataCase data, UnitSet nativeUnits) throws IOException {
        
        try {
            //  Skip down to the start of the array data.
            String aLine = in.readLine();
            while ((aLine != null) && (!aLine.startsWith("    1     TIME"))) {
                aLine = in.readLine();
            }
            if (aLine == null)
                throw new IOException("Could not find the 1st data line in CRF file.");
            
            //  Time is always the 1st parameter.
            ArrayParam arr = (ArrayParam)data.get(0);
            arr.setName("TIME");
            arr.setUserObject(RESOURCES.getString("ttSimTimeDesc"));
            arr = arr.changeTo(nativeUnits.time());
            data.set(0, arr);
 
            //  Define a set of dictionaries for each type of data that could be in the DAT file.
            HashMap<String,List<String>> sensorData = new HashMap();
            HashMap<String,List<String>> functionData = new HashMap();
            HashMap<String,List<String>> actuatorData = new HashMap();
            HashMap<String,List<String>> deviceData = new HashMap();
            HashMap<String,List<String>> controllerOutData = new HashMap();
            HashMap<String,List<String>> controllerInData = new HashMap();
            HashMap<String,List<String>> thetaData = new HashMap();
            HashMap<String,List<String>> thDotData = new HashMap();
            HashMap<String,List<String>> htransData = new HashMap();
            HashMap<String,List<String>> hvelData = new HashMap();
            HashMap<String,List<String>> hcontFData = new HashMap();
            HashMap<String,List<String>> hcontMData = new HashMap();
            HashMap<String,List<String>> hconstFData = new HashMap();
            HashMap<String,List<String>> hconstMData = new HashMap();
            HashMap<String,List<String>> wjiData = new HashMap();
            HashMap<String,List<String>> qbodyData = new HashMap();
            HashMap<String,List<String>> rvectData = new HashMap();
            HashMap<String,List<String>> massData = new HashMap();
            HashMap<String,List<String>> cmData = new HashMap();
            HashMap<String,List<String>> moipoiData = new HashMap();
            HashMap<String,List<String>> etaData = new HashMap();
            HashMap<String,List<String>> etaDotData = new HashMap();
            HashMap<String,List<String>> udjbjData = new HashMap();
            HashMap<String,List<String>> udotjData = new HashMap();
            HashMap<String,List<String>> psiEtaData = new HashMap();
            HashMap<String,List<String>> psiEtaDotData = new HashMap();
            
            //  Loop over all the columns defined in the CRF file.
            int col = 1;
            aLine = in.readLine();
            while (aLine != null && !"".equals(aLine)) {
                //  Split the line using repeated spaces as a delimiter.
                String[] parts = aLine.trim().split("\\s+");
                
                //  Parse out the column number.
                try {
                    String colStr = parts[0];
                    col = Integer.parseInt(colStr) - 1;
                } catch (NumberFormatException ex) {
                    throw new NoColumnNumberException();
                }
                
                //  Get the output type string.
                String type = parts[1];
                
                //  Get the array for the column in the DAT file we are working with.
                arr = (ArrayParam)data.get(col);
                switch (type) {
                    case "RSG":
                        //  We have a function generator output.
                        arr = parseCRFElementOutput(parts, arr, 0, functionData, FUNCTION, nativeUnits);
                        break;
                    
                    case "RC":
                        //  We have a controller output.
                        arr = parseCRFElementOutput(parts, arr, 0, controllerOutData, CONTROLLER_OUT, nativeUnits);
                        break;
                        
                    case "UC":
                        //  We have a controller input column.
                        arr = parseCRFElementOutput(parts, arr, 0, controllerInData, CONTROLLER_IN, nativeUnits);
                        break;
                        
                    case "RP":
                        //  We have a sensor output.
                        arr = parseCRFElementOutput(parts, arr, 0, sensorData, SENSOR, nativeUnits);
                        break;
                    
                    case "UP":
                        //  We have an actuator output.
                        arr = parseCRFDestInput(parts, arr, actuatorData, ACTUATOR);
                        break;
                        
                    case "DEMAG":
                        //  We have device output.
                        arr = parseCRFElementOutput(parts, arr, 0, deviceData, DEVICE, nativeUnits);
                        break;
                    
                    case "THETA":
                        //  We have hinge angle output.
                        arr = parseCRFElementOutput(parts, arr, 0, thetaData, HINGE, nativeUnits);
                        break;
                        
                    case "THDOT":
                        //  We have hinge angular rate output.
                        arr = parseCRFElementOutput(parts, arr, 3, thDotData, HINGE, nativeUnits);
                        break;
                        
                    case "WJI":
                        //  We have angular body rate outputs.
                        arr = parseCRFElementOutput(parts, arr, 0, wjiData, BODY, nativeUnits);
                        break;
                        
                    case "YJI":
                        //  We have hinge translation outputs.
                        arr = parseCRFElementOutput(parts, arr, 6, htransData, HINGE, nativeUnits);
                        break;
                        
                    case "YDOTJI":
                    case "YDJI":
                        //  We have hinge velocity outputs.
                        arr = parseCRFElementOutput(parts, arr, 9, hvelData, HINGE, nativeUnits);
                        break;
                        
                    case "ETA":
                        //  We have generalized modal coordinate outputs.
                        arr = parseCRFElementOutput(parts, arr, 3, etaData, "ETA", nativeUnits);
                        break;
                        
                    case "ETADOT":
                        //  We have generalized modal velocity outputs.
                        arr = parseCRFElementOutput(parts, arr, 3, etaDotData, "ETADOT", nativeUnits);
                        break;
                        
                    case "UDJBJ":
                        //  We have flex body deformation outputs.
                        arr = parseCRFElementOutput(parts, arr, 3, udjbjData, "UDJBJ", nativeUnits);
                        break;
                        
                    case "UDOTJ":
                        //  We have flex body deformation rate outputs.
                        arr = parseCRFElementOutput(parts, arr, 3, udotjData, "UDOTJ", nativeUnits);
                        break;
                        
                    case "PSIETA":
                        //  We have flex body rotational deformation outputs.
                        arr = parseCRFElementOutput(parts, arr, 3, psiEtaData, "PSIETA", nativeUnits);
                        break;
                        
                    case "PSETAD":
                        //  We have flex body rotation rate outputs.
                        arr = parseCRFElementOutput(parts, arr, 3, psiEtaDotData, "PSETAD", nativeUnits);
                        break;
                        
                    case "QBODY":
                        //  We have a TREETOPS quaternion output.
                        arr = parseCRFElementOutput(parts, arr, 3, qbodyData, BODY, nativeUnits);
                        break;
                        
                    case "Q_BODY":
                        //  We have a CLVTOPS quaternion output.
                        arr = parseCRFElementOutput(parts, arr, 7, qbodyData, BODY, nativeUnits);
                        break;
                        
                    case "RHJ":
                        //  We have hinge location vector output.
                        arr = parseCRFElementOutput(parts, arr, 11, rvectData, BODY, nativeUnits);
                        break;
                        
                    case "MASS":
                        //  We have body mass output.
                        arr = parseCRFElementOutput(parts, arr, 14, massData, BODY, nativeUnits);
                        break;
                       
                    case "CM":
                        //  We have body center of mass outputs.
                        arr = parseCRFElementOutput(parts, arr, 15, cmData, BODY, nativeUnits);
                        break;
                       
                    case "MOIPOI":
                        //  We have body inertia outputs.
                        arr = parseCRFElementOutput(parts, arr, 18, moipoiData, BODY, nativeUnits);
                        break;
                        
                    case "FCONI":
                        //  We have contact force outputs.
                        arr = parseCRFElementOutput(parts, arr, 12, hcontFData, HINGE, nativeUnits);
                        break;
                        
                    case "MCONI":
                        //  We have contact moment outputs.
                        arr = parseCRFElementOutput(parts, arr, 15, hcontMData, HINGE, nativeUnits);
                        break;
                        
                    case "FCJI":
                        //  We have constraint force outputs.
                        arr = parseCRFElementOutput(parts, arr, 18, hconstFData, HINGE, nativeUnits);
                        break;
                        
                    case "MCHJI":
                        //  We have constraint moment outputs.
                        arr = parseCRFElementOutput(parts, arr, 21, hconstMData, HINGE, nativeUnits);
                        break;
                        
                    case "NSTEPS":
                        //  We have the number of steps completed column.
                        arr.setName("NSteps");
                        arr.setUserObject(RESOURCES.getString("ttNStepsDesc"));
                        break;
                        
                    case "UNSUN":
                        //  We have solar direction unit vector components.
                        parseCRFSunUnitVector(parts, arr);
                        break;
                        
                    default:
                        //  We have an unknown type.
                        aLine = aLine.trim();
                        aLine = aLine.substring(aLine.indexOf(" ")).trim();
                        arr.setUserObject(aLine);
                        break;
                }
                
                //  Replace the array in the DataSet with the one with updated units.
                data.set(col, arr);
                
                //  Read in the next line.
                aLine = in.readLine();
            }
            
        } catch (Throwable ex) {
            throw new IOException("Error parsing CRF file on line #" + in.getLineNumber() + ".", ex);
        }
    }
 
    /**
     * Parse model element output line from the CRF file to get the element ID number and
     * output index.
     *
     * @param tokens       An array of tokens parsed from a line of CRF file input.
     * @param arr          The array of data from the data file for this line of input.
     * @param idxOffset    The offset into the list of outputs from the element for this
     *                     "type" of output.
     * @param existingData A dictionary containing data on the elements of a particular type
     *                     that have already been read in.
     * @param type         A constant representing the type of the data.
     * @param nativeUnits  The unit set representing the native units for this TreeTops
     *                     model.
     */
    private static ArrayParam parseCRFElementOutput(String[] tokens, ArrayParam arr, int idxOffset,
            HashMap<String, List<String>> existingData, String type, UnitSet nativeUnits) throws IOException, NumberFormatException {
        //  Get the output index number or element type of this type of item.
        String outIdxStr = tokens[2];
        
        //  Get the element ID.
        String idStr = "?";
        if (tokens.length > 3) {
            idStr = tokens[3];
            
            //  Make sure it is actually a number (an exception is thrown if it is not).
            Integer.parseInt(idStr);
        }
        
        //  Determine the index of the output in this column.
        int outputIdx;
        if (!"?".equals(idStr)) {
            List<String> existing = existingData.get(idStr);
            if (existing == null) {
                existing = new ArrayList();
                existingData.put(idStr, existing);
            }
            existing.add(idStr);
            outputIdx = existing.size() - 1 + idxOffset;    //  Make the index 0 offset.
        } else
            outputIdx = Integer.parseInt(outIdxStr) - 1;    //  Make index 0 offset.
        
        String desc, var;
        Unit unit = Dimensionless.UNIT;
        switch (type) {
            case BODY:
            case HINGE: {
                //  Get the output description for this output.
                desc = OUT_DESC.get(type)[outputIdx];
                //  Get the output variable name.
                var = OUT_VARS.get(type)[outputIdx];
                //  Get the output unit type.
                String unitType = OUT_UNITS.get(type)[outputIdx];
                //  Get the output units from the unit type and this model's native units.
                unit = getUnitFromType(nativeUnits, unitType);
                break;
            }
            case "ETA": {
                //  Skip the MODE column (idx = 4)
                //  Get the MODE ID #.
                String mIDStr = tokens[5].trim();
                desc = MessageFormat.format(RESOURCES.getString("ttFlexBodyModeIDDesc"),mIDStr);
                var = "ETA_" + mIDStr;
                type = BODY;
                break;
            }
            case "UDJBJ": {
                //  Skip the NODE column (idx = 4).
                //  Get the NODE ID #.
                String nIDStr = tokens[5].trim();
                //  Skip the AXIS column (idx = 6).
                //  Get the Axis ID#.
                String axisIDStr = tokens[7].trim();
                int axis = Integer.parseInt(axisIDStr);
                desc = MessageFormat.format(RESOURCES.getString("ttFlexBodyDispDesc"),nIDStr);
                var = "n" + nIDStr + ":UDJBJ_" + (axis == 1 ? "X" : (axis == 2 ? "Y" : "Z"));
                unit = nativeUnits.length();
                type = BODY;
                break;
            }
            case "UDOTJ": {
                //  Skip the NODE column (idx = 4).
                //  Get the NODE ID #.
                String nIDStr = tokens[5].trim();
                //  Skip the AXIS column (idx = 6).
                //  Get the AXIS ID #.
                String axisIDStr = tokens[7].trim();
                int axis = Integer.parseInt(axisIDStr);
                desc = MessageFormat.format(RESOURCES.getString("ttFlexBodyDispRateDesc"),nIDStr);
                var = "n" + nIDStr + ":UDOTJ_" + (axis == 1 ? "X" : (axis == 2 ? "Y" : "Z"));
                unit = nativeUnits.velocity();
                type = BODY;
                break;
            }
            case "PSIETA": {
                //  Skip the NODE column (idx = 4).
                //  Get the NODE ID #.
                String nIDStr = tokens[5].trim();
                //  Skip the AXIS column (idx = 6).
                //  Get the AXIS ID #.
                String axisIDStr = tokens[7].trim();
                int axis = Integer.parseInt(axisIDStr);
                desc = MessageFormat.format(RESOURCES.getString("ttFlexBodyRotDesc"),nIDStr);
                var = "n" + nIDStr + ":PSIETA_" + (axis == 1 ? "X" : (axis == 2 ? "Y" : "Z"));
                unit = SI.RADIAN;
                type = BODY;
                break;
            }
            case "PSETAD": {
                //  Skip the NODE column (idx = 4).
                //  Get the NODE ID #.
                String nIDStr = tokens[5].trim();
                //  Skip the AXIS column (idx = 6).
                //  Get the AXIS ID #.
                String axisIDStr = tokens[7].trim();
                int axis = Integer.parseInt(axisIDStr);
                desc = MessageFormat.format(RESOURCES.getString("ttFlexBodyRotRateDesc"),nIDStr);
                var = "n" + nIDStr + ":PSETAD_" + (axis == 1 ? "X" : (axis == 2 ? "Y" : "Z"));
                unit = AngularVelocity.UNIT;
                type = BODY;
                break;
            }
            case CONTROLLER_IN:
                desc = MessageFormat.format(RESOURCES.getString("ttControllerInputDesc"), outputIdx + 1);
                var = "Input" + (outputIdx + 1);
                break;
            case CONTROLLER_OUT:
                desc = MessageFormat.format(RESOURCES.getString("ttControllerOutputDesc"), outputIdx + 1);
                var = "Output" + (outputIdx + 1);
                break;
            default:
                desc = MessageFormat.format(RESOURCES.getString("ttOutputDesc"), outputIdx + 1);
                var = "Output" + (outputIdx + 1);
                break;
        }
        
        //  Construct a name from the information about this column.
        StringBuilder name = new StringBuilder();
        name.append(IDABR.get(type));
        name.append(":");
        name.append(idStr);
        name.append(":");
        name.append(var);
        
        //  Set the name, units, and reference data.
        arr.setName(name.toString());
        arr = arr.changeTo(unit);
        arr.setUserObject(desc);
        
        return arr;
    }

    /**
     * Parse an interconnect destination input line from the CRF file to get the element
     * ID number and output index.
     *
     * @param tokens       An array of tokens parsed from a line of CRF file input.
     * @param arr          The array of data from the DAT file for this line of input.
     * @param existingData A dictionary containing data on the destinations that have
     *                     already been read in.
     * @param type         A constant representing the type of the data.
     */
    private static ArrayParam parseCRFDestInput(String[] tokens, ArrayParam arr, 
            HashMap<String,List<String>> existingData, String type) {

        //  The output index number or element type.
        String inputIdxStr = tokens[2].trim();

        //  Get the element ID.
        String idStr = "?";
        if (tokens.length > 3) {
            idStr = tokens[3].trim();
            
            //  Make sure it is actually a number (an exception thrown if it is not).
            Integer.parseInt(idStr);
        }

        //  Determine the index of the input in this column.
        int inputIdx;
        if (!"?".equals(idStr)) {
            List<String> existing = existingData.get(idStr);
            if (existing == null) {
                existing = new ArrayList();
                existingData.put(idStr, existing);
            }
            existing.add(idStr);
            inputIdx = existing.size() - 1; //  Make index 0 offset.

        } else
            inputIdx = Integer.parseInt(inputIdxStr) - 1;   //  Make index 0 offset.

        //  Get the description for this input.
        String desc = MessageFormat.format(RESOURCES.getString("ttInputParDesc"), inputIdx + 1);

        //  Get the input variable name.
        String var = "Input" + (inputIdx + 1);

        //  Get the input units from the unit type and this model's native units.
        Unit unit = Dimensionless.UNIT;

        //  Construct a name from the information about this column.
        StringBuilder name = new StringBuilder();
        name.append(IDABR.get(type));
        name.append(":");
        name.append(idStr);
        name.append(":");
        name.append(var);

        //  Set the name, units, and reference data.
        arr.setName(name.toString());
        arr = arr.changeTo(unit);
        arr.setUserObject(desc);
        
        return arr;
    }

    /**
     * Parse a Sun unit vector line from the CRF file.
     *
     * @param tokens An array of tokens parsed from a line of CRF file input.
     * @param arr    The array of data from the DAT file for this line of input.
     */
    private static void parseCRFSunUnitVector(String[] tokens, ArrayParam arr)
            throws NoSuchElementException, NumberFormatException {
        //  Skip the "INERTIAL", "COMPONENTS," and "AXIS" tokens (idx = 2, 3 & 4).

        //  Get the axis index number.
        String axisStr = tokens[5].trim();
        int axis = Integer.parseInt(axisStr) - 1;       //  Make zero offset index.

        String[] axisOptions = {"USX", "USY", "USZ"};

        //  Construct a list of information about this column.
        StringBuilder name = new StringBuilder();
        name.append("UNSUN");
        name.append(":");
        name.append(axisOptions[axis]);

        //  Set the name and reference data.
        arr.setName(name.toString());
        arr.setUserObject(RESOURCES.getString("ttSunUVDesc"));
    }

    /**
     * Method that returns the unit from the provided unit set of the specified type.
     */
    private static Unit getUnitFromType(UnitSet unitSet, String type) throws IOException {
        //  Deal with some special cases first.
        if (null == type)
            return Dimensionless.UNIT;
        else switch (type) {
            case "nd":
                return Dimensionless.UNIT;
            case "angularVelocityR":
                return AngularVelocity.UNIT;
            case "angleR":
                return SI.RADIAN;
            default:
                break;
        }
        
        //  Use reflection to get the unit type from the unit set.
        //  "type" is the name of the method in unitSet to call.
        Unit unit = Dimensionless.UNIT;
        try {
            //  Get the appropriate method from unitSet.
            Method method = unitSet.getClass().getMethod( type, (Class[])null );
            
            //  Get the unit of the specified type.
            unit = (Unit)method.invoke(unitSet, (Object[])null);
            
        } catch (NoSuchMethodException | IllegalAccessException | InvocationTargetException e) {
            throw new IOException(e);
        }
        
        return unit;
    }
    
    /**
     * Sets the default set name to use.
     * 
     * @param name The name to use as the default set name.
     */
    @Override
    public void setDefaultSetName(CharSequence name) {
        datReader.setDefaultSetName(name);
    }

    /**
     * This class can not write to the CLVTOPS format file(s), so this method always throws
     * an exception.
     * 
     * @param parent
     * @param data
     * @return 
     */
    @Override
    public List<DataSet> selectDataForSaving(Frame parent, List<DataSet> data) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    /**
     * Method that writes out the data stored in the specified {@link DataSet} object to
     * the specified output stream in CLVTOPS format. This method will throw an exception
     * as writing to this format is not supported.
     *
     * @param output The output stream to which the data is to be written.
     * @param data   A list of {@link DataSet} objects containing data to be written out.
     * @throws IOException If there is a problem writing to the specified stream.
     */
    @Override
    public void write(OutputStream output, List<DataSet> data) throws IOException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    /**
     * A simple exception to catch the case of the column number missing in the CRF file.
     */
    @SuppressWarnings("serial")
    private static class NoColumnNumberException extends Exception {

        public NoColumnNumberException() {
        }
    }

}
