/**
 * EditUnitSetDialog -- A dialog that allows the user to select a set of units to be used.
 *
 * Copyright (C) 2008-2024, by Joseph A. Huwaldt. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA. Or visit: http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.unit;

import jahuwaldt.swing.EscapeJDialog;
import java.awt.Component;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.util.ResourceBundle;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;

/**
 * A modal dialog the allows the user to select a set of units to be used.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: July 21, 2008
 * @version March 27, 2024
 */
@SuppressWarnings("serial")
public class EditUnitSetDialog extends EscapeJDialog implements ActionListener {

    //  The resource bundle for this class.
    private static final ResourceBundle RESOURCES = EditUnitSetPanel.RESOURCES;
    
    //  The panel the user interacts with to select the units.
    private final EditUnitSetPanel unitPanel;

    //  The UnitSet being edited.
    private UnitSet outputUnits = null;

    /**
     * Construct a model dialog that allows the user to select a set of units to be used.
     *
     * @param parent  A reference to the parent component that this dialog belongs to
     *                (null is fine).
     * @param title   The title for the dialog window.
     * @param message A message to display telling the user what to do.
     * @param unitSet The unit set being edited.
     */
    @SuppressWarnings({"LeakingThisInConstructor","OverridableMethodCallInConstructor"})
    public EditUnitSetDialog(Component parent, String title, String message, UnitSet unitSet) {
        super(parent, title, true);

        //  Layout the dialog window.
        Container cp = this.getContentPane();
        cp.setLayout(new BoxLayout(cp, BoxLayout.Y_AXIS));

        //  Add the edit panel to this dialog.
        unitPanel = new EditUnitSetPanel(message, unitSet, true);
        cp.add(unitPanel);

        //  Define OK and Cancel buttons.
        Box panel = Box.createHorizontalBox();
        cp.add(panel);
        panel.setBorder(BorderFactory.createEmptyBorder(4, 4, 4, 4));
        panel.add(Box.createGlue());

        JButton cancelBtn = new JButton(RESOURCES.getString("cancel"));
        cancelBtn.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                performEscapeAction(null);
            }
        });
        panel.add(cancelBtn);

        panel.add(Box.createHorizontalStrut(40));

        JButton okayBtn = new JButton(RESOURCES.getString("okay"));
        okayBtn.addActionListener(this);
        this.getRootPane().setDefaultButton(okayBtn);
        panel.add(okayBtn);

        panel.add(Box.createHorizontalStrut(20));

        //  Pack the window.
        this.pack();

        //  Position the dialog to be near the parent window.
        this.setLocationRelativeTo(parent);
}

    /**
     * Returns the unit set that has been edited by the user.
     *
     * @return The unit set that has been edited by the user or null if the user cancels
     *         the edit.
     */
    public UnitSet getUnits() {
        return outputUnits;
    }

    /**
     * This method is called when the user clicks on the "OK Button".
     *
     * @param evt The event that caused this method to be called (ignored).
     */
    @Override
    public void actionPerformed(ActionEvent evt) {

        //  Set the output units to be the units we have been editing.
        outputUnits = unitPanel.getUnits();

        //  Close the window.
        this.dispatchEvent(new WindowEvent(this, WindowEvent.WINDOW_CLOSING));
    }

}
