/**
 * EditUnitSetPanel -- A panel that allows the user to select a set of units to be used.
 *
 * Copyright (C) 2008-2024, by Joseph A. Huwaldt. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA. Or visit: http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.unit;

import jahuwaldt.swing.JButtonGroup;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.measure.converter.ConversionException;
import javax.measure.unit.Unit;
import javax.swing.*;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;

/**
 * A panel the allows the user to select a set of units to be used.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: October 14, 2008
 * @version March 27, 2024
 */
@SuppressWarnings("serial")
public class EditUnitSetPanel extends JPanel {

    /**
     * The resource bundle for use in this class.
     */
    static final ResourceBundle RESOURCES = ResourceBundle.getBundle("jahuwaldt.js.unit.EditUnitSetResources", java.util.Locale.getDefault());
    
    //  The column names in the derived units table.
    private static final String[] derColumnNames = {
        RESOURCES.getString("unitTypeHeaderLabel"), 
        RESOURCES.getString("unitValueHeaderLabel")
    };

    //  The combo-boxes that contain the fundamental units.
    private JComboBox timeCBox, lengthCBox, massCBox;

    //  The data stored in the derived units table.
    private final Object[][] derRowData = new Object[16][2];

    //  The JTable displaying the derived units.
    private final JTable derTable;

    //  The UnitSet being edited.
    private UnitSet theUnits = null;

    /**
     * Construct a panel that allows the user to select a set of units to be used.
     *
     * @param message    A message to display telling the user what to do.
     * @param unitSet    The unit set being edited.
     * @param showCustom The option for custom units is shown if this is true.
     */
    @SuppressWarnings("OverridableMethodCallInConstructor")
    public EditUnitSetPanel(String message, UnitSet unitSet, boolean showCustom) {
        super();

        theUnits = (UnitSet)unitSet.clone();

        //  Create the data to display in the derivative units table.
        setDerivedData();

        //  Layout the panel.
        this.setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
        this.setAlignmentX(Component.CENTER_ALIGNMENT);

        //  Add instructions at the top.
        Box topPanel = Box.createHorizontalBox();
        topPanel.setAlignmentX(Component.RIGHT_ALIGNMENT);
        this.add(topPanel);

        topPanel.add(Box.createHorizontalStrut(10));
        topPanel.add(new JLabel(message));

        //  Create a panel to display in the center.
        Box centerPanel = Box.createVerticalBox();
        centerPanel.setBorder(BorderFactory.createEmptyBorder(10, 20, 10, 20));
        this.add(centerPanel);

        //  Create a set of radio buttons for selecting the unit set to use.
        Box panel = Box.createVerticalBox();
        panel.setAlignmentX(Component.CENTER_ALIGNMENT);
        centerPanel.add(panel);
        panel.setBorder(BorderFactory.createTitledBorder(RESOURCES.getString("coherentUnitSetsTitle")));
        JButtonGroup btnGroup = new JButtonGroup();

        JRadioButton rb = new JRadioButton(RESOURCES.getString("siUnitSetTxt"));
        panel.add(rb);
        boolean found = false;
        if (unitSet.equals(new UnitSet(UnitSet.UnitSystem.SI_MKS))) {
            rb.setSelected(true);
            found = true;
        }
        rb.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                theUnits = new UnitSet(UnitSet.UnitSystem.SI_MKS);
                updateDerivativesTable();
                setEnableFundamentalUnitBoxes(false);
                setFundamentalUnitSelections();
            }
        });
        btnGroup.add(rb);

        rb = new JRadioButton(RESOURCES.getString("cgsUnitSetTxt"));
        panel.add(rb);
        if (!found && unitSet.equals(new UnitSet(UnitSet.UnitSystem.CGS))) {
            rb.setSelected(true);
            found = true;
        }
        rb.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                theUnits = new UnitSet(UnitSet.UnitSystem.CGS);
                updateDerivativesTable();
                setEnableFundamentalUnitBoxes(false);
                setFundamentalUnitSelections();
            }
        });
        btnGroup.add(rb);

        rb = new JRadioButton(RESOURCES.getString("fssUnitSetTxt"));
        panel.add(rb);
        if (!found && unitSet.equals(new UnitSet(UnitSet.UnitSystem.US_FSS))) {
            rb.setSelected(true);
            found = true;
        }
        rb.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                theUnits = new UnitSet(UnitSet.UnitSystem.US_FSS);
                updateDerivativesTable();
                setEnableFundamentalUnitBoxes(false);
                setFundamentalUnitSelections();
            }
        });
        btnGroup.add(rb);

        rb = new JRadioButton(RESOURCES.getString("issUnitSetTxt"));
        panel.add(rb);
        if (!found && unitSet.equals(new UnitSet(UnitSet.UnitSystem.US_ISS))) {
            rb.setSelected(true);
            found = true;
        }
        rb.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                theUnits = new UnitSet(UnitSet.UnitSystem.US_ISS);
                updateDerivativesTable();
                setEnableFundamentalUnitBoxes(false);
                setFundamentalUnitSelections();
            }
        });
        btnGroup.add(rb);

        rb = new JRadioButton(RESOURCES.getString("fpsUnitSetTxt"));
        panel.add(rb);
        if (!found && unitSet.equals(new UnitSet(UnitSet.UnitSystem.US_FPS))) {
            rb.setSelected(true);
            found = true;
        }
        rb.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                theUnits = new UnitSet(UnitSet.UnitSystem.US_FPS);
                updateDerivativesTable();
                setEnableFundamentalUnitBoxes(false);
                setFundamentalUnitSelections();
            }
        });
        btnGroup.add(rb);

        if (showCustom) {
            rb = new JRadioButton(RESOURCES.getString("customUnitSetTxt"));
            panel.add(rb);
            if (!found)
                rb.setSelected(true);
            rb.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    setEnableFundamentalUnitBoxes(true);
                }
            });
            btnGroup.add(rb);

            //  Create a series of combo-boxes that allow the user to choose custom combinations.
            centerPanel.add(Box.createVerticalStrut(10));
            panel = Box.createHorizontalBox();
            panel.setBorder(BorderFactory.createTitledBorder(RESOURCES.getString("fundamentalUnitOptionsLabel")));
            centerPanel.add(panel);

            panel.add(new JLabel(RESOURCES.getString("timeLabel") + "=", JLabel.RIGHT));
            timeCBox = new JComboBox(UnitSet.getSet(UnitSet.SetType.TIME));
            timeCBox.setMaximumSize(timeCBox.getPreferredSize());
            timeCBox.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    JComboBox cbox = (JComboBox)e.getSource();
                    int value = cbox.getSelectedIndex();
                    if (value >= 0) {
                        handleUnitChange(0, (Unit)cbox.getSelectedItem());
                    }
                }
            });
            panel.add(timeCBox);

            panel.add(new JLabel(RESOURCES.getString("lengthLabel") + "=", JLabel.RIGHT));
            lengthCBox = new JComboBox(UnitSet.getSet(UnitSet.SetType.LENGTH));
            lengthCBox.setMaximumSize(lengthCBox.getPreferredSize());
            lengthCBox.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    JComboBox cbox = (JComboBox)e.getSource();
                    int value = cbox.getSelectedIndex();
                    if (value >= 0) {
                        handleUnitChange(1, (Unit)cbox.getSelectedItem());
                    }
                }
            });
            panel.add(lengthCBox);

            panel.add(new JLabel(RESOURCES.getString("massLabel") + "=", JLabel.RIGHT));
            massCBox = new JComboBox(UnitSet.getSet(UnitSet.SetType.MASS));
            massCBox.setMaximumSize(massCBox.getPreferredSize());
            massCBox.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    JComboBox cbox = (JComboBox)e.getSource();
                    int value = cbox.getSelectedIndex();
                    if (value >= 0) {
                        handleUnitChange(2, (Unit)cbox.getSelectedItem());
                    }
                }
            });
            panel.add(massCBox);
        }

        //  Create a display of the derived units.
        centerPanel.add(Box.createVerticalStrut(10));
        panel = Box.createHorizontalBox();
        panel.setBorder(BorderFactory.createTitledBorder(RESOURCES.getString("derivedUnitsLabel")));
        centerPanel.add(panel);

        //  Create a table to display the derived units.
        derTable = new JTable(new DerivativeTableModel());
        derTable.setCellSelectionEnabled(false);
        derTable.setColumnSelectionAllowed(false);
        derTable.setRowSelectionAllowed(false);
        derTable.setCellEditor(null);
        derTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);     //  Show scroll bars.

        //  Initialize the column widths to show all the data.
        initColumnWidths(derTable);

        //  Determine the table preferred width.
        int width = 0;
        int numColumns = derTable.getColumnCount();
        for (int i = 0; i < numColumns; ++i) {
            TableColumn column = derTable.getColumnModel().getColumn(i);
            width += column.getPreferredWidth();
        }
        derTable.setPreferredScrollableViewportSize(new Dimension(width, 110));

        //  Put the table in a scroll pane so that we have scroll bars.
        JScrollPane scrollPane = new JScrollPane(derTable);
        scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
        scrollPane.setMaximumSize(new Dimension(width + 20, 250));

        //  Add the list to the panel.
        panel.add(scrollPane);

        //  Set the intially selected items.
        setFundamentalUnitSelections();

        //  Enable/disable the unit boxes as appropriate.
        setEnableFundamentalUnitBoxes(!found);
    }

    /**
     * Method that handles the user changing one of the fundamental unit types. This is
     * called whenever the user selects a new unit.
     */
    private void handleUnitChange(int idx, Unit newUnit) {

        try {

            switch (idx) {
                case 0:     //  Change in time units.
                    if (!newUnit.equals(theUnits.time()))
                        theUnits.setTime(newUnit);
                    break;

                case 1:     //  Change in length units.
                    if (!newUnit.equals(theUnits.length()))
                        theUnits.setLength(newUnit);
                    break;

                case 2:     //  Change in mass units.
                    if (!newUnit.equals(theUnits.mass()))
                        theUnits.setMass(newUnit);
                    break;
            }

            //  Create a coherent unit set.
            theUnits.makeConsistent();

            //  Update the derived units table.
            updateDerivativesTable();

        } catch (ConversionException ex) {
            Logger.getLogger(EditUnitSetPanel.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * Method that updates the derived unit table display.
     */
    private void updateDerivativesTable() {
        //  Change the derived table data set.
        setDerivedData();

        //  Notify the derived table that it's data has changed.
        AbstractTableModel model = (AbstractTableModel)derTable.getModel();
        model.fireTableDataChanged();
    }

    /**
     * Enabled or disable all the fundamental unit combo-boxes.
     */
    private void setEnableFundamentalUnitBoxes(boolean enable) {
        if (timeCBox != null)
            timeCBox.setEnabled(enable);
        if (lengthCBox != null)
            lengthCBox.setEnabled(enable);
        if (massCBox != null)
            massCBox.setEnabled(enable);
    }

    /**
     * Set the currently selected item in the fundamental unit boxes.
     */
    private void setFundamentalUnitSelections() {
        if (timeCBox != null)
            timeCBox.setSelectedItem(theUnits.time());
        if (lengthCBox != null)
            lengthCBox.setSelectedItem(theUnits.length());
        if (massCBox != null)
            massCBox.setSelectedItem(theUnits.mass());
    }

    /**
     * Method that fills in the derived row data table with the current units.
     */
    private void setDerivedData() {
        int row = 0;
        derRowData[row][0] = RESOURCES.getString("areaLabel");      derRowData[row++][1] = theUnits.area();
        derRowData[row][0] = RESOURCES.getString("volumeLabel");    derRowData[row++][1] = theUnits.volume();
        derRowData[row][0] = RESOURCES.getString("forceLabel");     derRowData[row++][1] = theUnits.force();
        derRowData[row][0] = RESOURCES.getString("inertiaLabel");   derRowData[row++][1] = theUnits.inertia();
        derRowData[row][0] = RESOURCES.getString("densityLabel");   derRowData[row++][1] = theUnits.massDensity();
    }

    /**
     * Returns the unit set stored with this panel.
     *
     * @return The unit set stored with this panel.
     */
    public UnitSet getUnits() {
        return theUnits;
    }

    /**
     * Method that resets the minimum column widths.
     */
    private static void initColumnWidths(JTable table) {
        int numColumns = table.getColumnCount();
        for (int i = 0; i < numColumns; ++i) {
            //  Get each column.
            TableColumn column = table.getColumnModel().getColumn(i);

            //  Determine the header width.
            int headerWidth = 0;
            try {
                TableCellRenderer renderer = column.getHeaderRenderer();
                if (renderer == null)
                    renderer = table.getTableHeader().getDefaultRenderer();

                Component comp = renderer.
                        getTableCellRendererComponent(
                                null, column.getHeaderValue(),
                                false, false, 0, 0);
                headerWidth = comp.getPreferredSize().width;
            } catch (NullPointerException e) {
                e.printStackTrace();
            }

            if (i == 0)
                column.setPreferredWidth(headerWidth + 10);
            else
                //  Set the minimum width to the larger of the header width
                //  and 100 pixels.
                column.setPreferredWidth(Math.max(100, headerWidth) + 10);
        }
    }

    /**
     * Table model for the derivative unit table.
     */
    private class DerivativeTableModel extends AbstractTableModel {

        @Override
        public String getColumnName(int col) {
            return derColumnNames[col];
        }

        @Override
        public int getRowCount() {
            return derRowData.length;
        }

        @Override
        public int getColumnCount() {
            return derColumnNames.length;
        }

        @Override
        public Object getValueAt(int row, int col) {
            return derRowData[row][col];
        }

        @Override
        public boolean isCellEditable(int row, int col) {
            return false;
        }

        @Override
        public Class getColumnClass(int c) {
            return getValueAt(0, c).getClass();
        }
    }
}
