/*
 *  qPlot  -- A Java Application for quickly plotting tabulated data.
 *
 *  Copyright (C) 2011-2023, Joseph A. Huwaldt. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */

import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.MDIApplication;
import jahuwaldt.swing.MainApp;
import jahuwaldt.swing.Preferences;
import java.awt.Frame;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;
import javax.measure.unit.SI;
import javax.measure.unit.UnitFormat;

/**
 * Application to make quick interactive plots of tabulated data.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: September 3, 2011
 * @version December 9, 2023
 */
public class qPlot implements MainApp {

    /**
     * The resource bundle containing the string resources for this application. Resource
     * bundles are useful for localizing applications. New localities (languages) can be
     * added by adding properties files only -- no code modifications.
     */
    private static final ResourceBundle RESBUNDLE = ResourceBundle.getBundle("appStrings", Locale.getDefault());
    
    //	A reference to our GUI application instance.
    private final MDIApplication guiApp;
    
    /**
     * The user preferences for this program.
     */
    private final Preferences prefs;

    /**
     * Entry point for this application. This is where it starts up from.
     * @param args  An array of command line arguments
     */
    public static void main(final String[] args) {

        // Schedule a job for the event-dispatching thread:
        // creating and showing this application's GUI.
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                try {

                    //  Define some alternate unit aliases.
                    UnitFormat.getInstance().alias(SI.SECOND, "sec");
                            
                    // Create an instance of our application.
                    if (args.length > 0) {
                        new qPlot(args);
                    } else {
                        new qPlot();
                    }

                } catch (Exception e) {
                    e.printStackTrace();
                    AppUtilities.showException(null, RESBUNDLE.getString("unexpectedTitle"),
                            RESBUNDLE.getString("unexpectedMsg"), e);
                    System.exit(0);
                }
            }
        });

    }

    /**
     * Constructor for our application that displays the GUI.
     * @throws java.lang.Exception
     */
    @SuppressWarnings("LeakingThisInConstructor")
    public qPlot() throws Exception {

        guiApp = new AppGUI(RESBUNDLE);

        //	Load in the user preferences.
        prefs = new AppPreferences();
        MDIApplication.setPreferences(prefs);

        // Finally, create an instance of an application window to get the program rolling.
        Frame window = MainWindow.newAppWindow(null);
        window.setVisible(true);
        MDIApplication.addWindow(window);
    }

    /**
     * Constructor for our application that reads in a plot definition file from the command line.
     * @param args  An array of command line arguments.
     * @throws java.lang.Exception
     */
    @SuppressWarnings("LeakingThisInConstructor")
    public qPlot(String[] args) throws Exception {

        guiApp = new AppGUI(RESBUNDLE);

        //	Load in the user preferences.
        prefs = new AppPreferences();
        MDIApplication.setPreferences(prefs);

        //	Deal with the command line arguments.

        //	Return a list of arguments without the options.
        List<String> argsNoOptions = argumentsWithoutOptions(args);

        if (argumentStartsWith(args, "-h") || argumentStartsWith(args, "--help"))
            printHelpAndDie();

        if (argumentStartsWith(args, "-v")) {
            System.out.print("qPlot version ");
            System.out.println(RESBUNDLE.getString("appVersion"));
            System.out.println();
        }

        int defIndex = argumentIndex(args, "--defs");
        File defsFile = null;
        if (defIndex >= 0)
            defsFile = new File(args[defIndex + 1]);

        //	Create an instance of the application window using the supplied data file.
        if (!argsNoOptions.isEmpty()) {
            List<File> files = new ArrayList<>();
            for (String filename : argsNoOptions) {
                File file = new File(filename);
                files.add(file);
            }
            Frame window = MainWindow.newWindowFromDataFile(new Frame(), files, true, defsFile);
            //  The window is made visible automatically.
            MDIApplication.addWindow(window);

        } else
            System.exit(0);
    }

    /**
     * Returns true if any of the arguments starts with the specified string (ignoring
     * case).
     */
    private boolean argumentStartsWith(String[] args, String value) {
        value = value.toLowerCase();
        int length = args.length;
        for (int i = 0; i < length; ++i) {
            if (args[i].toLowerCase().startsWith(value))
                return true;
        }
        return false;
    }

    /**
     * Takes the input list of arguments and strips out any arguments that are options
     * (i.e.: start with "-").
     */
    private List<String> argumentsWithoutOptions(String[] args) {
        List<String> list = new ArrayList<>();
        int length = args.length;
        for (int i = 0; i < length; ++i) {
            if (!args[i].startsWith("-"))
                list.add(args[i]);

            //	Skip the next argument if this one starts with "--".
            if (args[i].startsWith("--"))
                ++i;
        }
        return list;
    }

    /**
     * Return the index to the argument that starts with the specified value (ignoring
     * case).
     *
     * @param args The list of all the arguments.
     * @param value The value we are looking for.
     * @return the index to the argument that starts with the specified value ignoring
     *      case or -1 if none could be found.
     */
    private int argumentIndex(String[] args, String value) {
        value = value.toLowerCase();
        int length = args.length;
        for (int i = 0; i < length; ++i) {
            if (args[i].toLowerCase().startsWith(value))
                return i;
        }
        return -1;
    }

    //-----------------------------------------------------------------------------------
    /**
     * Method that prints a help message to standard out and then exits the program.
     */
    private static void printHelpAndDie() {
        System.out.println();
        System.out.println("qPlot -- Quick Plotting Program");
        System.out.println("Usage:");
        System.out.println("    java -jar qPlot.jar [--defs plotDefsFile] filename");
        System.out.println("    where:");
        System.out.println("        plotDefsFile = The optional plot definition file to use for pre-defined plots,");
        System.out.println("        filename = The data file to be plotted.");
        System.out.println();
        System.exit(0);
    }

    /**
     * Returns the resource bundle stored with this application. If no resource bundle has
     * been stored, then null is returned.
     * @return the resource bundle stored with this application.
     */
    @Override
    public final ResourceBundle getResourceBundle() {
        return RESBUNDLE;
    }

    /**
     * Return a reference to this program's GUI application or null if we are running in
     * batch mode.
     * @return a reference to this program's GUI application or null if we are running in
     * batch mode.
     */
    @Override
    public final MDIApplication getGUIApplication() {
        return guiApp;
    }

    /**
     * Return a reference to the user preferences for this application.
     * @return a reference to the user preferences for this application.
     */
    @Override
    public final Preferences getPreferences() {
        return prefs;
    }
}
